"""
This module is a implementation of the miner present by R. Agrawal in 1997.

See the following article for more information about the agrawal miner:\n
R. Agrawal, D. Gunopulos, and F. Leymann, “Mining process models from workflow
logs,” in EDBT, ser. Lecture Notes in Computer Science, vol. 1377, 
Springer, 1998, pp. 469–483.
"""
from dataclasses import dataclass, field
from typing import Set, Optional, List, Tuple, Dict
from copy import deepcopy
from itertools import product
from logging import DEBUG, INFO
from warnings import warn

from pmkoalas.discovery.meta import DiscoveryTechnique
from pmkoalas.simple import Trace
from pmkoalas.simple import EventLog
from pmkoalas._logging import debug, enable_logging, info, get_logger

@dataclass
class DependencyNode():
    """
    Helper
    """

    label:str

    def __hash__(self) -> int:
        return hash(self.label)
    
    def __eq__(self, value: object) -> bool:
        if isinstance(value, type(self)) :
            return self.__hash__() == value.__hash__()
        else:
            False


@dataclass
class DependencyEdge():
    """
    Helper
    """

    source:DependencyNode
    target:DependencyNode
    counter:int=1

    def increment(self) -> None:
        self.counter += 1

    def __hash__(self) -> int:
        return hash((self.source.label,self.target.label))
    
    def __eq__(self, value: object) -> bool:
        if isinstance(value, type(self)) :
            return self.__hash__() == value.__hash__()
        else:
            return False
    
@dataclass
class DependencyWalk():

    walk:List[DependencyNode] = field(default_factory=list)


    def step(self, edge:DependencyEdge) -> 'DependencyWalk':
        last = self.walk[-1]

        if (edge.source == last):
            contains = [ n for n in self.walk if n == edge.target] 
        else:
            raise ValueError("Passed a edge that is not from the last node.")


class DependencyGraph():
    """
    This graph represents the concept of a dependency graph,
    that is, a graph that represents all the dependencies found in the log,
    as described by R. Agrawal et al. in 1998 in def. 5.

    The vertices of the graph are the seen activities in the log, 
    and the edges denote that activity v depends on activity u (def. 4).

    The graph contains one starting vertex and one ending vertex.
    """

    def __init__(self, 
                 vertices:Set[DependencyNode], edges:Set[DependencyEdge],
                 _reachability:Dict[DependencyNode,Dict[DependencyNode,bool]]=None,
                 ignore_start_end:bool=False,
                ) -> None:
        if (len(vertices) == 0 and len(edges) != 0):
            raise ValueError("Cannot have edges without vertices.")
        self._vertices = set()
        self._edges = set()
        self._start = None
        self._end = None 
        self._inedges = dict()
        self._outedges = dict()
        if _reachability == None:
            self._reachability = dict()
        else:
            self._reachability = deepcopy(_reachability)
        # check that each vertex is in at least one edge
        remove = set()
        for v in vertices:
            vedges = [ e for e in edges 
                      if e.source == v or e.target == v]
            if len(vedges) == 0:
                warn("DependencyGraph was given a vertex without edges, " 
                     +"removing.")
                continue
            # find start and end vertices
            tgt = [e for e in edges if e.target == v]
            src = [e for e in edges if e.source == v]
            self._inedges[v] = set(tgt)
            self._outedges[v] = set(src)
            if (v not in self._reachability):
                self._reachability[v] = dict()
            if len(tgt) == 0:
                if (self._start != None):
                    if (not ignore_start_end):
                        raise ValueError("Cannot have multiple start vertices.")
                self._start = v
            if len(src) == 0:
                if (self._end != None):
                    if (not ignore_start_end):
                        raise ValueError("Cannot have multiple end vertices.")
                self._end = v
            # store vertex and edges
            self._vertices.add(deepcopy(v))
            for e in vedges:
                self._edges.add(deepcopy(e))
        # query observed activities 
        self._acts = set([ v.label for v in self._vertices])
        # only keep the ordered walk of vertices
        # walkv = self.walk_vertices()
        # walke = self.walk_edges()
        if (self._start == None and len(vertices) > 0):
            raise ValueError("Missing start vertex.")
        if (self._end == None and len(vertices) > 0):
            raise ValueError("Missing end vertex.")

    def is_empty(self) -> bool:
        """
        Returns True if the graph is empty.
        """
        return len(self._vertices) == 0
    
    def start(self) -> DependencyNode:
        """
        Returns the start vertex of the graph.
        """
        return deepcopy(self._start)
    
    def end(self) -> DependencyNode:    
        """
        Returns the end vertex of the graph.
        """
        return deepcopy(self._end)
    
    def vertices(self) -> Set[DependencyNode]:
        """
        Returns the vertices of the graph as a new set.
        """
        return deepcopy(self._vertices)
    
    def edges(self) -> Set[DependencyEdge]:
        """
        Returns the edges of the graph as a new set.
        """
        return deepcopy(self._edges)

    def induce_subgraph(self, trace:Trace) -> 'DependencyGraph':
        """
        Returns the induced subgraph of the dependency graph for the given
        trace, may be an empty graph.
        """
        acts = trace.seen_activities()
        retV = set([ v for v in self.vertices() if v.label in acts])
        pedges = set([ e for e in self.edges() if e.source in retV])
        retE = set([ 
            e 
            for e 
            in pedges
            if _check_follows_for(trace, e.target.label, e.source.label)
        ])
        return DependencyGraph(
            retV,
            retE,
            ignore_start_end=True    
        )
        
    def isconnected(self) -> bool:
        """
        Returns True if the graph is connected.
        """
        if self.is_empty():
            return False

        # starting point
        start_vertex = self.start()
        debug(f"starting from vertex ::  {start_vertex.label}")

        # Perform depth-first search starting from the start_vertex
        visited = set()
        stack = [start_vertex]

        while len(stack) > 0:
            vertex = stack.pop()
            visited.add(vertex)
            debug(f"visiting {vertex.label}")

            # Get the neighbors of the current vertex
            neighbors = [ e.target for e in self.outedges(vertex) ]

            # Add unvisited neighbors to the stack
            unvisited_neighbors = [neighbor for neighbor in neighbors if neighbor not in visited]
            stack.extend(unvisited_neighbors)

        # If all vertices have been visited, the graph is connected
        debug(f"visited the following vertices :: {visited}")
        debug(f"len check :: {len(visited)} vs {len(self._vertices)}")
        return len(visited) == len(self._vertices) and self.end() in visited
    
    def path_between(self, from_node:DependencyNode, to_node:DependencyNode
                    ) -> bool:
        """
        Returns True if there is a path from from_node to to_node in the graph.
        """
        # check if the path has already been calculated
        past = self._reachability[from_node]
        if to_node in past:
            return past[to_node]
        # compute new reachability
        visited = set()
        stack = [from_node]

        while stack:
            current_node = stack.pop()
            visited.add(current_node)

            if current_node == to_node:
                self._reachability[from_node][to_node] = True
                return True

            neighbors = [ e.target for e in self.outedges(current_node) ]
            unvisited_neighbors = [neighbor for neighbor in neighbors if neighbor not in visited]
            stack.extend(unvisited_neighbors)

        self._reachability[from_node][to_node] = False
        return False
    
    def does_not_violate(self, trace:Trace) -> bool:
        """
        Returns if the given trace does not violate any dependency.

        Ideally, should only be called on the induced subgraph of the 
        trace.
        """
        # check the starting 
        start = trace[0]
        if start != self.start().label:
            return False
        # check the ending
        end = trace[-1]
        if end != self.end().label:
            return False
        # now check the body of the trace
        for act in trace.seen_activities():
            deps = [ e for e in self.edges() if e.target.label == act]
            for edge in deps:
                if not _check_follows_for(trace, edge.target.label, edge.source.label):
                    return False
        # alll cases held, so the trace does not violate
        return True
    
    def transitive_reduction(self) -> 'DependencyGraph':
        """
        Returns the transitive reduction of the graph.
        """
        # create a copy of the graph
        retV = deepcopy(self.vertices())
        retE = deepcopy(self.edges())
        # calculate reachability for each pair of vertices
        for u in self.vertices():
            for v in self.vertices():
                self.path_between(u, v)
        reachability = self._reachability
        # reduce edges until only the smallest subset of edges remain that 
        # ensures that all connected members remain connected. 
        for v in self.vertices():
            for u in self.vertices():
                if u != v:
                    for w in self.vertices():
                        if w != u and w != v:
                            if reachability[u][v] and reachability[v][w] and reachability[u][w]:
                                debug(f"removing edge {u.label} -> {w.label}")
                                retE.discard(DependencyEdge(u, w))
        return DependencyGraph(retV, retE, ignore_start_end=True)
    
    def outedges(self, node:DependencyNode) -> Set[DependencyEdge]:
        """
        Returns the outedges of the given node.
        """
        return sorted( 
            self._outedges[node],
            key=lambda e: e.target.label
        )
    
    def get_neighbors(self, node:DependencyNode) -> Set[DependencyNode]:
        """
        Returns the neighbors of the given node.
        """
        return sorted(
            set([ e.target for e in self.outedges(node)]),
            key=lambda n: n.label
        )
    
    def walk_edges(self) -> List[DependencyEdge]:
        """
        Returns a consistence walk of the edges in the graph.
        """
        ret = [ ]
        stack = [ self.start() ]
        seen = set()
        while len(stack) > 0:
            node = stack.pop()
            seen.add(node)
            for edge in self.outedges(node):
                if edge not in seen:
                    ret.append(edge)
                    seen.add(edge)
                if (edge.target not in seen):
                    stack.append(edge.target)
        assert len(ret) == len(self.edges())
        return ret

    def walk_vertices(self) -> List[DependencyNode]:
        """
        Returns a consistence walk of the vertices in the graph.
        """
        ret = [ self.start() ]
        stack = [ self.start() ]
        seen = set()
        while len(stack) > 0:
            node = stack.pop()
            seen.add(node)
            if node not in ret and node != self.end():
                ret.append(node)
            for edge in self.outedges(node):
                if edge.target not in seen:
                    stack.append(edge.target)
        ret.append(self.end())
        assert len(ret) == len(self.vertices())
        return ret
    
    def create_dot_form(self) -> str:
        ret = "digraph{"
        ret += "dpi=150;rankdir=LR;nodesep=0.6;ranksep=0.3;"
        ret += "node[shape=circle,fillcolor=lightgray,style=filled,width=2,penwidth=8,fontsize=18,fontname=\"roboto\"];"
        ret += "edge[penwidth=4,fontsize=16,minlen=2,fontname=\"roboto\"];"
        nodeIds = dict()
        nid = 0
        for v in self.walk_vertices():
            nodeIds[v] = f"n{nid}"
            label = '<'+'<BR/>'.join([v.label[i:i+16] for i in range(0, len(v.label), 16)])+">"
            if v == self.start():
                ret += f"{nodeIds[v]}[label={label},fillcolor=green];" 
            elif v == self.end():
                ret += f"{nodeIds[v]}[label={label},fillcolor=red];" 
            else:
                ret += f"{nodeIds[v]}[label={label}];"
            nid += 1
        assert nid == len(self._vertices)
        elabel = 0
        seen = set()
        for v in self.walk_vertices():
            for e in self.outedges(v):
                if e.target not in seen:
                    ret += f"d{elabel}[shape=none,color=black,width=0.5,"\
                        +  f"label={e.counter},style=none];"
                    ret += f"{nodeIds[e.source]} -> d{elabel}:w"\
                        + f"[arrowhead=odot,color=black];"
                    ret += f"d{elabel}:e -> {nodeIds[e.target]}"\
                        + f"[arrowtail=odot,dir=both,color=black];"
                else:
                    ret += f"d{elabel}[shape=none,color=black,width=0.5,"\
                        +  f"label={e.counter},style=none];"
                    ret += f"{nodeIds[e.target]} -> d{elabel}:w"\
                        + f"[arrowhead=odot,dir=both,color=gray];"
                    ret += f"d{elabel}:e -> {nodeIds[e.source]}"\
                        + f"[arrowtail=odot,dir=back,color=gray];"
                elabel += 1
            seen.add(v)
        ret += "}"
        assert elabel == len(self._edges)
        return ret
    
    # model access functions
    
    def __eq__(self, value: object) -> bool:
        if issubclass(type(value), (DependencyGraph)):
            return (
                self.vertices() == value.vertices()
                and self.edges() == value.edges()
            )
        else:
            return False
        
    def __repr__(self) -> str:
        ret = "DependencyGraph(\n"
        ret += "\tset([\n"
        for v in self.walk_vertices():
            ret += f"\t\t{v.__repr__()},\n"
        ret += "\t]),\n"
        ret += "\tset([\n"
        for e in self.walk_edges():
            ret += f"\t\tDependencyEdge(\n"
            ret += f"\t\t\t{e.source.__repr__()},\n"
            ret += f"\t\t\t{e.target.__repr__()},\n"
            ret += f"\t\t\t{e.counter},\n"
            ret += "\t\t),\n"
        ret += "\t]),\n"
        ret += ")"
        return ret
        
    
def _check_follows_for(trace:Trace, actB:str, actA:str) -> bool:
        """
        Returns True if the first instance of activity A occurs before 
        the first instance activity B in the given trace.
        """
        first_a = -1
        first_b = -1
        for id,act in enumerate(trace.__iter__()):
            if act == actA:
                first_a = id
            if act == actB:
                first_b = id
            if first_a != -1 and first_b != -1:
                break
        debug(f"checking :: {trace} for {actA} -> {actB} :: {first_a} < {first_b}")
        return first_a < first_b

def compute_follows_relations(slog:EventLog) -> Set[Tuple[str,str]]:
        """
        Returns the set of follows relations for the given trace based on
        def.3 in Agrawal et al. 1998. That is each entry in the returned
        set (B,A) denotes that activity B follows activity A.

        An acitivty B follows activity A if B starts after A in each 
        trace that both B and A appear in, 
        OR,
        there exists an activity C such that C follows A and B follows C.
        """
        debug(slog)
        acts = slog.seen_activities()
        debug(acts)
        follows = set()
        # iterate over all traces and compute case 1
        for actB,actA in product(acts,acts):
            if actA == actB:
                continue
            debug(f"checking {actB} -> {actA}")
            # collect all executions where both acts occur
            traces = [ 
                trace 
                for trace,_ 
                in slog.__iter__() 
                if 
                    actA in trace.seen_activities() 
                    and 
                    actB in trace.seen_activities()
            ]
            if (len(traces) == 0):
                continue
            # check if B follow A in all traces
            debug(traces)
            traces = [
                1 if _check_follows_for(t,actB,actA) else 0
                for t 
                in traces
            ]
            debug(traces)
            if (sum(traces) == len(traces)):
                follows.add((actB,actA))
        # adjust follows based on case 2
        addtions = set()
        for actB,actA in product(acts,acts):
            if actB == actA:
                continue
            if (actB,actA) in follows:
                continue
            for actC in acts.difference({actB,actA}):
                debug(f"checking that {actC} -> {actA}, and {actB} -> {actC}")
                if (actC,actA) in follows and (actB,actC) in follows:
                    debug(f"found that {actB} follows {actA} through {actC}")
                    addtions.add((actB,actA))
                    break
        debug(follows)
        debug(addtions)
        follows = follows.union(addtions)
        return follows
    
def compute_dependencies(slog:EventLog) -> Set[Tuple[str,str]]:
    """
    Returns the set of dependencies for the given event log.
    That is a set, where each entry (B,A) denotes that activity 
    B depends on activity A.
    """
    follows = compute_follows_relations(slog)
    acts = slog.seen_activities()
    ret = set()
    # check for an activity A and activity B,
    # that B follows A, but A does not follow B,
    # then B depends on A.
    for actB,actA in product(acts,acts):
        if actB == actA:
            continue
        if (actB,actA) in follows and (actA,actB) not in follows:
            debug(f"found that {actB} depends on {actA}")
            ret.add((
                actB,
                actA
            ))
    return ret

def compute_independent_activities(slog:EventLog) -> Set[Tuple[str,str]]:
    """
    Returns the set of independent activities for the given event log.
    That is a set, where each entry (B,A) denotes that activity A and
    activity B are indpendent.
    """
    follows = compute_follows_relations(slog)
    acts = slog.seen_activities()
    ret = set()
    # check for an activity A and activity B,
    # such that either B follows A, but A follows B,
    # or B does not follow A, but A does not follow B.
    for actB,actA in product(acts,acts):
        if actB == actA:
            continue
        if (actB,actA) in follows and (actA,actB) in follows:
            debug(f"found that {actB} and {actA} are independent (both)")
            ret.add((
                actB,
                actA
            ))
            continue
        if (actB,actA) not in follows and (actA,actB) not in follows:
            debug(f"found that {actB} and {actA} are independent (neither)")
            ret.add((
                actB,
                actA
            ))
    return ret

def execution_is_consistent(trace:Trace, graph:DependencyGraph) -> bool:
    """
    Determines if the trace is consistent with the given graph, as described
    by definition 6 in R. Agrawal, D. Gunopulos, and F. Leymann, “Mining 
    process models from workflow logs”.
    """
    # check that the seen activities are a subset of the verticies 
    if (trace.seen_activities().issubset(graph._acts) == False):
        return False
    try:
        sub = graph.induce_subgraph(trace)
    except ValueError:
        return False

    if (sub.isconnected()):
        # now check that there exists a path between start and all 
        # other vertices
        start = sub.start()
        for v in sub._vertices:
            if (v == start):
                continue
            if (sub.path_between(start,v) == False):
                return False
        # check that no dependecies are violated by the ordering 
        # in the trace
        return sub.does_not_violate(trace)

    return False

def is_conformal(slog:EventLog, graph:DependencyGraph) -> bool:
    """
    Determines if the given event log is conformal with the given graph, as 
    described by definition 7 in R. Agrawal et. al. 1998.
    
    A dependency graph G is conformal with a log if the following holds:
        - for each dependency in  slog, there exists a path in G;
        - there is no path in G between independent activities;
        - G is consistent with every trace in the slog.
    """
    # check dependencies 
    for deps in compute_dependencies(slog):
        if not graph.path_between(
            DependencyNode(deps[1]), DependencyNode(deps[0])):
            return False
    # check that independent activities are not connected
    for indep in compute_independent_activities(slog):
        if graph.path_between(
            DependencyNode(indep[0]), DependencyNode(indep[1])):
            return False
    # check that the graph is consistent with every trace
    for trace,_ in slog.__iter__():
        if not execution_is_consistent(trace, graph):
            return False
    return True

def find_strongly_connected_components(graph: DependencyGraph) -> List[Set[DependencyNode]]:
    """
    Finds the strongly connected components in the given graph using Tarjan's algorithm.
    Returns a list of sets, where each set represents a strongly connected component.
    """
    index_counter = 0
    index = dict()
    low_link = dict()
    on_stack = set()
    stack = []
    components = []

    def strongconnect(node: DependencyNode):
        nonlocal index_counter, low_link, index, stack, on_stack, components
        index[node] = index_counter
        low_link[node] = index_counter
        index_counter += 1
        stack.append(node)
        on_stack.add(node)

        for neighbor in graph.get_neighbors(node):
            if neighbor not in index:
                strongconnect(neighbor)
                low_link[node] = min(low_link[node], low_link[neighbor])
            elif neighbor in on_stack:
                low_link[node] = min(low_link[node], index[neighbor])

        if low_link[node] == index[node]:
            component = set()
            while True:
                neighbor = stack.pop()
                on_stack.remove(neighbor)
                component.add(neighbor)
                if neighbor == node:
                    break
            components.append(component)

    for node in graph.vertices():
        if node not in index:
            strongconnect(node)

    return components

class AgrawalMinerInstance(DiscoveryTechnique):
    """
    Mines a graph from an simplified event log using the proposed 
    approach by R. Agrawal D. Gunopulos and F. Leymann in 1998.

    The resulting graph aims to have the following properties:
    - completeness, the graph preserves all dependencies between activities
      presented in the event log.
    - irredundancy, the graph should not introduce spurious dependencies 
    - mimimality, the graph should have miminal edges.

    The resulting graph represents the constraints between activities rather
    than a system that executes to produce traces.
    """

    def __init__(self, 
                 optimise_step_five:bool=False, 
                 min_instances:int=1) -> None:
        """
        Initialises the miner for the following calls of `discover'.
         
        Internal Parameters:\n
        `optimise_step_five` (`bool=False`) 
            which sets what version of step five should be used, sequential 
            marking (False) or parallel marking over traces (True).
        `min_instances` (`int=1`)
            which sets the number of times an edge must be observed to be
            considered.
        """
        self._use_opt_five = optimise_step_five
        self._min_instances = min_instances

    def _test_for_consistent_activity_usage(self, slog:EventLog, n:int) -> bool:
        """
        Returns True if the event log, consistently uses the same activities
        for traces in the log, where at exactly n occurances of said 
        activities within traces.
        """
        acts = slog.seen_activities()
        for trace,_ in slog.__iter__():
            tacts = trace.seen_activities()
            # check that all activities are observed
            if tacts < acts:
                debug(f"trace {trace} does not contain all activities")
                return False
            seen = dict()
            # check that activities are observed n times
            for act in trace:
                if act in seen:
                    seen[act] += 1
                else:
                    seen[act] = 1
                if (seen[act] > n):
                    debug(
                        f"trace {trace} contains activity {act} more than \
                        {n} times")
                    return False
            # check ret matches n
            for act in tacts:
                if seen[act] != n:
                    debug(
                        f"trace {trace} does not contain activity {act} \
                        {n} times")
                    return False
        return True
    
    def _test_for_no_duplicated_activities(self, slog:EventLog) -> bool:
        """
        Tests whether the traces in the log have no duplicated activities.
        """
        for trace,_ in slog.__iter__():
            seen = set()
            # check that activities are observed n times
            for act in trace:
                if act in seen:
                    return False
                else:
                    seen.add(act)
        return True 
    
    def _adjust_traces(self, trace:Trace, prefix:bool, suffix:bool) -> Trace:
        """
        Adds a start and end activity to the trace if required.
        """
        ret = list(trace)
        if prefix:
            ret = [ "start" ] + ret
        if suffix:
            ret = ret + [ "end" ]
        return Trace(ret)
    
    @enable_logging
    def discover(self, slog:EventLog) -> DependencyGraph:
        """
        Discovers the dependency graph from the given simplifed event log, 
        using the proposed approach by R. Agrawal D. Gunopulos and F. 
        Leymann in 1998.

        For discovery, the conceptualisation of a process requires a single
        source and a single sink. Thus, abstracted starts and ends are 
        introduced into traces (if required).

        One of three algorithms will be used to discover the graph:
            - the special DAG algorithm, if the log consistently uses the
                same set of activities for a trace (exactly once).
            - the general DAG algorithm, if the log samples from the same
                set of activities for a trace (without replacement).
            - the cyclic DAG algorithm, if the log samples from a process
                where activities could repeated in a trace.
        """
        # adjust traces as required
        adjust_traces = False
        prefix = False
        suffix = False
        if len(slog.seen_start_activities()) != 1 \
            or len(slog.seen_end_activities()) != 1:
            if len(slog.seen_start_activities()) != 1:
                prefix = True
            if len(slog.seen_end_activities()) != 1:
                suffix = True
            adjust_traces = True
        if adjust_traces:
            info("adjusting traces by adding start and end activities " 
                 +"as needed")
            slog = EventLog([
                self._adjust_traces(t,prefix,suffix)
                for t,n in slog.__iter__()
                for i in range(n)
            ])
        # decide on the algorithm to use
        if self._test_for_consistent_activity_usage(slog, 1):
            info("using the special DAG algorithm")
            return self._discover_special_dag(slog)
        else:
            if self._test_for_no_duplicated_activities(slog):
                info("using the general DAG algorithm")
                return self._discover_general_dag(slog)
            else:
                info("using the cyclic DAG algorithm")
                return self._discover_cyclic_dag(slog)

    def _discover_special_dag(self, slog:EventLog)-> DependencyGraph:
        """
        Discovers the special DAG from the given simplifed event log, 
        using the proposed approach by R. Agrawal D. et. al. in 1998 
        through algorithm 1.

        The algorithm consists of the following steps:
            - start with a graph consisting of vertices for each activity
                and no edges.
            - for each execution E and activities u,v in E, add the edge of
                (u, v) if u terminates before v starts.
            - remove all edges that appear in both directions, i.e. 
                for activities u,v, if the edges (u,v) and (v,u) are in 
                the graph, remove both of them.
            - Compute the transitive closure of the graph.
            - Return the graph.
        """
        retV = self._step_one(slog)
        # step two 
        retE = self._step_two(slog)
        # step three
        graph = self._step_three(slog,retV,retE)
        # step four 
        info("step four started")
        graph = graph.transitive_reduction()
        info("step four completed")
        # step five
        info("returning a graph (|V|,|E|) :: ({},{})".format(
            len(graph.vertices()),len(graph.edges()))
            )
        return graph
    
    def _discover_general_dag(self, slog:EventLog) -> DependencyGraph:
        """
        Discovers the general DAG from the given simplifed event log, 
        using the proposed approach by R. Agrawal D. et. al. in 1998 
        through algorithm 2.

        The algorithm consists of the following steps:
            - start with a graph consisting of vertices for each activity
                and no edges.
            - for each execution E and activities u,v in E, add the edge of
                (u, v) if u terminates before v starts.
            - remove all edges that appear in both directions, i.e. 
                for activities u,v, if the edges (u,v) and (v,u) are in 
                the graph, remove both of them.
            - For each strongly connected component of G, remove from E 
                all edges between vertices in the same strongly connected 
                component.
            - For each process execution in L:
                - Find the induced subgraph of G.
                - Compute the transitive reduction of the subgraph.
                - Mark those edges in E that are present in the transitive 
                    reduction.
            - Remove the unmarked edges in E.
            - Return the graph.
        """
        retV = self._step_one(slog)
        retE = self._step_two(slog)
        graph = self._step_three(slog,retV,retE)
        graph = self._step_four(graph,retV,retE)
        if self._use_opt_five:
            marked = self._step_five_opt(slog,graph)
        else:
            marked = self._step_five(slog,graph)
        # step six
        graph = self._step_six(marked,graph)
        # step seven
        info("returning a graph (|V|,|E|) :: ({},{})".format(
            len(graph.vertices()),len(graph.edges()))
            )
        return graph
    
    def _suffix_repeated_activities(self, trace:Trace) -> Trace:
        """
        Replaces repeated activities in the trace with a suffix.
        """
        index = dict()
        ret = []
        for act in trace:
            if act not in index:
                ret.append(act)
                index[act] = 0
            else:
                index[act] = index[act] + 1
                id = index[act]
                ret.append(f"{act}##{id:02d}")
        return Trace(ret)
    
    def _discover_cyclic_dag(self, slog:EventLog) -> DependencyGraph:
        """
        Discover the cyclic DAG from the given simplifed event log,
        using the proposed approach by R. Agrawal D. et. al. in 1998
        through algorithm 3.

        The algorithm consists of the following steps:
            - process the log and replace repeated activities with a suffix
            - start with a graph consisting of vertices for each activity
                and no edges.
            - for each execution E and activities u,v in E, add the edge of
                (u, v) if u terminates before v starts.
            - remove all edges that appear in both directions, i.e. 
                for activities u,v, if the edges (u,v) and (v,u) are in 
                the graph, remove both of them.
            - For each strongly connected component of G, remove from E 
                all edges between vertices in the same strongly connected 
                component.
            - For each process execution in L:
                - Find the induced subgraph of G.
                - Compute the transitive reduction of the subgraph.
                - Mark those edges in E that are present in the transitive 
                    reduction.
            - Remove the unmarked edges in E.
            - merge the vertices that correspond to the different instances
                of the same activity and introduce edges between the merged
                verticies, if an subinstance had a edge between unmerged 
                vertices.
            - Return the graph.
        """
        slog = EventLog([
            self._suffix_repeated_activities(t)
            for t,i
            in slog
            for n 
            in range(i)
        ])
        retV = self._step_one(slog)
        retE = self._step_two(slog)
        graph = self._step_three(slog,retV,retE)
        graph = self._step_four(graph,retV,retE)
        if self._use_opt_five:
            marked = self._step_five_opt(slog,graph)
        else:
            marked = self._step_five(slog,graph)
        # step six
        graph = self._step_six(marked,graph)
        # step seven
        graph = self._step_seven(graph)
        info("returning a graph (|V|,|E|) :: ({},{})".format(
            len(graph.vertices()),len(graph.edges()))
            )
        return graph
    
    def _step_one(self, slog:EventLog) -> Set[DependencyNode]:
        """
        Creates a vertex for each activity seen in the log.
        """
        info("step one started")
        acts = slog.seen_activities()
        # step one
        retV = set([
            DependencyNode(act) 
            for act 
            in slog.seen_activities()
        ])
        info("step one completed")
        return retV

    def _step_two(self, slog:EventLog) -> Set[DependencyEdge]:
        """
        Computes the set of edges between activities, where each edge (A,B)
        denotes that activity A was eventually followed by activity B.

        If min_instances is set, then the edge must be observed at least
        min_instances times, by default set to 1. 
        """
        if self._use_opt_five:
            return self._step_two_opt(slog)
        info("step two started")
        retE = set()
        edges = dict()
        for trace,n in slog.__iter__():
            for left,right in product(range(len(trace)-1),range(1,len(trace))):
                if (left >= right):
                    continue
                edge = DependencyEdge(
                    DependencyNode(trace[left]),
                    DependencyNode(trace[right]),
                    n
                )
                if edge in retE:
                    for i in range(n):
                        edges[edge].increment()
                else:
                    retE.add(edge)
                    edges[edge] = edge
        # filter edges based on min_instances
        drops = set()
        for edge in retE:
            if edge.counter < self._min_instances:
                drops.add(edge)
        retE = retE.difference(drops)
        info("step two completed")
        return retE
    
    def _step_two_opt(self, slog:EventLog) -> Set[DependencyEdge]:
        """
        Parallel version of step two, that computes all eventually follows
        edges.
        """
        info("step two started")
        retE = set()
        edges = dict()
        from joblib import Parallel, delayed
        from tqdm import tqdm
        pool = Parallel(n_jobs=-2,return_as="generator_unordered")
        # define work 
        def work(trace:Trace, n:int)\
            -> Set[DependencyEdge]:
            edges = set()
            for left, right in product(
                range(len(trace)-1), range(1, len(trace))):
                if (left >= right):
                    continue
                edges.add(DependencyEdge( 
                    DependencyNode(trace[left]),
                    DependencyNode(trace[right]),
                    n)
                )
            return edges
        # prep work 
        tasks = [(trace, n ) for trace, n in slog]
        tasks = pool(delayed(work)(*task) for task in tasks)
        # add pb if on INFO 
        if get_logger().level == INFO:
            tasks = tqdm(tasks, desc="Launching jobs for step two", 
                         total=slog.get_nvariants())
        # sync work
        for workee in tasks:
            for edge in workee:
                if edge in edges:
                    for i in range(edge.counter):
                        edges[edge].increment()
                else:
                    edges[edge] = edge
                    retE.add(edge)
        # filter edges based on min_instances
        drops = set()
        for edge in retE:
            if edge.counter < self._min_instances:
                drops.add(edge)
        retE = retE.difference(drops)
        info("step two completed")
        return retE
        
    
    def _step_three(self, slog:EventLog, 
                    retV:Set[DependencyNode], retE:Set[DependencyEdge])\
        -> DependencyGraph:
        """
        Removes all edges that appear in both directions.
        """
        info("step three started")
        acts = slog.seen_activities()
        if get_logger().level == DEBUG:
            graph = DependencyGraph(retV,retE)
            debug(graph.__repr__())
            debug("graph before step three :: "+ graph.create_dot_form())
        for actA in acts:
            for actB in acts.difference({actA}):
                one_way = DependencyEdge(
                    DependencyNode(actA),
                    DependencyNode(actB)
                )
                if one_way in retE:
                    other_way = DependencyEdge(
                        DependencyNode(actB),
                        DependencyNode(actA)
                    )
                    if other_way in retE:
                        retE.discard(one_way)
                        retE.discard(other_way)
        graph = DependencyGraph(retV,retE)
        if get_logger().level == DEBUG:
            debug("graph after step three :: "+ graph.create_dot_form())
        info("step three completed")
        return graph

    def _step_four(self, graph:DependencyGraph, 
                   retV:Set[DependencyNode], retE:Set[DependencyEdge])\
        -> DependencyGraph:
        """
        Finds strongly connected components in the graph and removes all
        edges between vertices in the same strongly connected component.
        """
        info("step four started")
        components = find_strongly_connected_components(graph)
        for comp in components:
            for u in comp:
                for v in comp:
                    if u != v:
                        retE.discard(DependencyEdge(u,v))
        graph = DependencyGraph(retV,retE)     
        if get_logger().level == DEBUG:
            debug("graph after step four :: "+ graph.create_dot_form())
        info("step four completed")  
        return graph

    def _step_five(self, slog:EventLog, graph:DependencyGraph)\
        -> Set[DependencyEdge]:
        """
        Finds and marks all edges that are present in the induced subgraph
        for each trace, after the transitive reduction has been computed.
        """
        info("step five started")
        marked = set()
        for trace,_ in slog:
            sub = graph.induce_subgraph(trace)
            sub = sub.transitive_reduction()
            for edge in sub.edges():
                marked.add(edge)
        info("step five completed")
        return marked
    
    def _step_five_opt(self, slog:EventLog, graph:DependencyGraph)\
        -> Set[DependencyEdge]:
        """
        Optimised version of step five, that marks edges in parallel for
        all traces.
        """
        info("step five (opt) started")
        from joblib import Parallel, delayed
        from tqdm import tqdm
        pool = Parallel(n_jobs=-2,return_as="generator_unordered")
        # the work
        def mark(trace:Trace, graph:str)\
            -> str:
            from pmkoalas.discovery.agrawal_miner import DependencyGraph
            from pmkoalas.discovery.agrawal_miner import DependencyEdge 
            from pmkoalas.discovery.agrawal_miner import DependencyNode
            graph:DependencyGraph = eval(graph)
            marked = set()
            graph = graph.induce_subgraph(trace)
            graph = graph.transitive_reduction()
            for edge in graph.edges():
                marked.add(edge)
            return graph.edges().__repr__()
        # run the work
        graph_state = graph.__repr__()
        marks = pool(
            delayed(mark)(trace,graph_state)
            for trace,_ in slog
        )
        if (get_logger().level == INFO):
            marks = tqdm(marks, desc="launching jobs for step five (opt)", 
                 total=slog.get_nvariants())
        # sync over sets and keep the union from work
        marked = set()
        for markee in marks:
            markee = eval(markee)
            marked = marked.union(markee)
        info("step five (opt) completed")
        return marked

    
    def _step_six(self, marked:Set[DependencyEdge], graph:DependencyGraph)\
        -> DependencyGraph:
        """
        Removed unmarked edges from the graph.
        """
        info("step six started")
        retV = graph.vertices()
        retE = graph.edges()
        retE = retE.intersection(marked)
        info("step six completed")
        return DependencyGraph(retV,retE)

    def _step_seven(self, graph:DependencyGraph) -> DependencyGraph:
        """
        Merge the vertices into groups of equivalent activities, adding 
        edges between the merged vertices if an edge existed between a 
        member of a group and a member of another group.
        """
        info("step seven started")
        # find the groups of equivalent activities
        if (get_logger().level == DEBUG):
            debug("graph before step seven :: "+ graph.create_dot_form())
        groups = dict()
        for v in graph.vertices():
            label = v.label.split("##")[0]
            if label in groups:
                groups[label].add(v)
            else:
                groups[label] = set([v])
        debug(f"|groups| :: {len(groups)}")
        debug(f"found groups :: {groups}")
        retV = set()
        retE = set()
        edges = dict()
        # recreate nodes and edges
        for label in groups.keys():
            retV.add(DependencyNode(label))
            # add edges from the outedges for each member
            for member in groups[label]:
                for edge in graph.outedges(member):
                    tgt = edge.target.label.split("##")[0]
                    src = edge.source.label.split("##")[0]
                    if tgt == src:
                        continue
                    edge = DependencyEdge(
                        DependencyNode(src),
                        DependencyNode(tgt),
                        edge.counter
                    )
                    if edge in edges:
                        for i in range(edge.counter):
                            edges[edge].increment()
                    else:
                        edges[edge] = edge
                        retE.add(edge)
        graph = DependencyGraph(retV,retE)
        info("step seven completed")
        if (get_logger().level == DEBUG):
            debug("graph after step seven :: "+ graph.create_dot_form())
        return graph

    def compute_follows_relations(self, slog:EventLog) -> Set[Tuple[str,str]]:
        return compute_follows_relations(slog)
    
    def compute_dependencies(self, slog:EventLog) -> Set[Tuple[str,str]]:
        return compute_dependencies(slog)
    
    def compute_indepentent_activities(self, slog:EventLog) -> Set[Tuple[str,str]]:
        return compute_independent_activities(slog)


"""
This module outlines a data structure for transitions tree as proposed by Hidders, J.,
 Dumas, M., van der Aalst, W.M.P., ter Hofstede, A.H.M., Verelst, J.: When are two
 workflows the same? In: CATS. ACS (2005); These trees can be modified for future work.
"""
from typing import Any, Set, List, Tuple, Union
from copy import deepcopy
from dataclasses import dataclass
from functools import reduce
from os import path,mkdir

from pmkoalas.simple import Trace, EventLog
from pmkoalas.complex import ComplexEvent, ComplexEventLog
from pmkoalas._logging import info, InfoQueueProcessor, InfoIteratorProcessor

class TransitionTreeVertex():
    """
    Data class for a vertex in a transition tree.
    """

    VERTEX_FORMAT = "v_{id}"

    def __init__(self, id:int, partial_trace:Trace, end:bool=False) -> None:
        self._label = self.VERTEX_FORMAT.format(id=id)
        self._partial = deepcopy(partial_trace)
        self._end = end
    
    # properties
    def id(self) -> str:
        " returns a identifier for this vertex."
        return self._label
    
    def sigma_sequence(self) -> Trace:
        " returns the unique partial trace from root to this vertex."
        return deepcopy(self._partial)
    
    def is_root(self) -> bool:
        " returns a boolean to say if this vertex is the root."
        return False
    
    def set_as_terminal(self):
        " makes vertex terminal."
        self._end = True
    
    def terminal(self) -> bool:
        """ 
        returns a boolean to say if this vertex is signals the end of the 
        process execution.
        """
        return self._end
    
    # utility functions
    def outgoing(self, flows:Set['TransitionTreeFlow']) -> Set['TransitionTreeFlow']:
        " returns the set of flows that are outgoing from this vertex."
        out = set()
        for f in flows:
            if (f.offering() == self):
                out.add(f)
        return out
    
    def html_label(self) -> str:
        "returns a html like label with suffix."
        id = self.id().split("_")[-1]
        return f"<v<sub>{id}</sub>>"
    
    # data model functions
    def __str__(self) -> str:
        return self.id()

    def __eq__(self, __o: object) -> bool:
        if (isinstance(__o, TransitionTreeVertex)):
            return self.__hash__() == __o.__hash__()
        return False
    
    def __hash__(self) -> int:
        return hash((self.sigma_sequence().__hash__()))
    
class TransitionTreeRoot(TransitionTreeVertex):
    """
    Data class for the root of a transition tree.
    """

    def __init__(self, id=1) -> None:
        super().__init__(id, Trace([]))

    def is_root(self) -> bool:
        return True
    
    def html_label(self) -> str:
        "returns a html like label with suffix."
        return "<v<sub>r</sub>>"
    
class TransitionPlayoutRoot(TransitionTreeRoot):
    """
    Data class for the root of playout transition tree.
    """

    def html_label(self) -> str:
        return "< &lt;&gt; >"
    
class TransitionPlayoutVertex(TransitionTreeVertex):

    def html_label(self) -> str:
        acts = ""
        for act in self.sigma_sequence():
            acts += act+","
        acts = acts[:-1]
        return f"< &lt;{acts}&gt; >"

class TransitionTreeFlow():
    """
    Template for a flow between two vertices in a transition tree.
    """

    def __init__(self, source:TransitionTreeVertex, 
                 act:str, 
                 target:TransitionTreeVertex) -> None:
        self._source = source
        self._act = act 
        self._target = target

    def offering(self) -> TransitionTreeVertex:
        " returns the vertex that offers this flow."
        return self._source
    
    def next(self) -> TransitionTreeVertex:
        " returns the vertex that this flow is directed towards."
        return self._target
    
    def activity(self) -> str:
        " returns the process activity related to this flow."
        return self._act
    
    def label(self) -> object:
        " returns the actual label of the flow, varys depending on implementation."
        return self.activity()
    
    def attributes(self) -> Set[str]:
        """ 
        returns the associated attributes with this flow, varys depending on 
        implementation. 
        """
        return set()
    
    # data model functions
    def __str__(self):
        return f"{self.offering()} -> {self.activity()} -> {self.next()}"

    def __eq__(self, __o: object) -> bool:
        if (isinstance(__o, TransitionTreeFlow)):
            return self.__hash__() == __o.__hash__()
        return False
    
    def __hash__(self) -> int:
        return hash(tuple([ self.offering(), self.activity(), self.next()]))
    
class TransitionTreeInformationFlow(TransitionTreeFlow):
    """
    A implementation for a flow with abstracted information.
    """

    def __init__(self, source:TransitionTreeVertex, 
                 act:str, 
                 target:TransitionTreeVertex,
                 information:object) -> None:
        super().__init__(source, act, target)
        self._info = deepcopy(information)

    def label(self) -> Tuple[str, object]:
        return (self.activity(), deepcopy(self._info))

class TransitionTreePopulationFlow(TransitionTreeFlow):
    """
    A implementation for a flow with an exact population of data attribute mappings.
    """

    def __init__(self, source:TransitionTreeVertex, 
                 act:str, 
                 target:TransitionTreeVertex, 
                 population:List[ComplexEvent]) -> None:
        super().__init__(source, act, target)
        self._pop = deepcopy(population)
        self._attrs = set()
        for ev in self._pop:
            for attr in ev.data().keys():
                self._attrs.add(attr)

    def label(self) -> Tuple[str, List[ComplexEvent]]:
        return (self.activity(), deepcopy(self._pop))
    
    def attributes(self) -> Set[str]:
        return self._attrs
    
    def population(self) -> List[ComplexEvent]:
        " returns the associated population of data attribute mappings."
        return deepcopy(self._pop)
    
    def __str__(self):
        return "(" + super().__str__() + f")^{len(self._pop)}"
    
class TransitionTreeGuard():
    """
    A template for semantic of a guard from the system.
    """

    def __init__(self) -> None:
        pass

    def check(self, data:ComplexEvent) -> bool:
        " return whether the given data is supported by this guard."
        return False
    
    def required(self) -> Set[str]:
        " returns the required data attributes for this guard."
        return set()
    
    def enabling(self, population:Set[ComplexEvent]) -> Set[ComplexEvent]:
        " returns the subset of the population that this guard supports."
        return set([ p for p in population if self.check(p)])
    
    def html_label(self) -> str:
        return f"g<sub>1</sub>"
    
    def __hash__(self) -> int:
        return hash(self.html_label())
    
    def __eq__(self, other: object) -> bool:
        if (type(self) == type(other)):
            return self.__hash__() == other.__hash__()
        return False
    
class TransitionTreeMerge(TransitionTreeGuard):
    """
    A template for the semantics of merging two guards.
    """

    MERGER = """<sub><font point-size="22">&#8855;</font></sub>"""

    def __init__(self, left:TransitionTreeGuard, right:TransitionTreeGuard) -> None:
        super().__init__()
        self._contains = set()
        if isinstance(left, TransitionTreeMerge):
            self._contains = self._contains.union(left.contains())
        else:
            self._contains.add(left)
        if isinstance(right, TransitionTreeMerge):
            self._contains = self._contains.union(right.contains())
        else:
            self._contains.add(right)

    def contains(self) -> Set[TransitionTreeGuard]:
        return deepcopy(self._contains)

    def check(self, data: ComplexEvent) -> bool:
        check = True 
        for exp in self._contains:
            check = check and exp.check(data)
        return check
    
    def required(self) -> Set[str]:
        ret = set()
        for exp in self._contains:
            ret = ret.union(exp.required())
        return ret
    
    def html_label(self) -> str:
        ret = ""
        for exp in self._contains:
            ret += exp.html_label()
            ret += " " + self.MERGER + " "
        drop = -1 * (len(self.MERGER) + 1)
        ret = ret[:drop]
        return ret

    def __str__(self) -> str:
        return self.html_label()
    
    def __hash__(self) -> int:
        glabels = list( exp.html_label() for exp in self._contains)
        glabels = [self.MERGER] + sorted(glabels)
        return hash(tuple(glabels))
    
    def __eq__(self, other: object) -> bool:
        if (type(self) == type(other)):
            return self.__hash__() == other.__hash__()
        return False
    
class TransitionTreeJoin(TransitionTreeGuard):
    """
    A template for the semantics of joining to two guards.
    """

    JOIN = """<sub><font point-size="22">&#8854;</font></sub>"""

    def __init__(self, flows:Set['TransitionTreeGuardFlow']) -> None:
        super().__init__()
        self._contains = set( f.guard() for f in flows)

    def consists_of(self) -> Set[TransitionTreeGuard]:
        return deepcopy(self._contains)
    
    def check(self, data: ComplexEvent) -> bool:
        check = False 
        for exp in self._contains:
            check = check or exp.check(data)
        return check
    
    def required(self) -> Set[str]:
        ret = set()
        for exp in self._contains:
            ret = ret.union(exp.required())
        return ret
    
    def html_label(self) -> str:
        ret = ""
        for exp in self._contains:
            ret += "(" + exp.html_label() +")"
            ret += " " + self.JOIN + " "
        drop = -1 * (len(self.JOIN) + 1)
        ret = ret[:drop]
        return ret

    def __str__(self) -> str:
        return self.html_label()
    
    def __hash__(self) -> int:
        glabels = list( exp.html_label() for exp in self._contains)
        glabels = [self.JOIN] + sorted(glabels)
        return hash(tuple(glabels))
    
    def __eq__(self, other: object) -> bool:
        if (type(self) == type(other)):
            return self.__hash__() == other.__hash__()
        return False

class TransitionTreeGuardFlow(TransitionTreeFlow):
    """
    A implemenatiton for a flow with an implicit population using a guard 
    function.
    """

    def __init__(self, source:TransitionTreeVertex, 
                 act:str, 
                 target:TransitionTreeVertex, 
                 guard:TransitionTreeGuard) -> None:
        super().__init__(source, act, target)
        self._guard = guard

    def label(self) -> Tuple[str, TransitionTreeGuard]:
        return (self.activity(), self._guard)
    
    def attributes(self) -> Set[str]:
        return self._guard.required()
    
    def guard(self) -> TransitionTreeGuard:
        return self._guard
    
    def __hash__(self) -> int:
        return hash(tuple([
            self.offering(),
            hash(self.activity())+self.guard().__hash__(),
            self.next()
        ]))
    
    def __str__(self):
        return f"{self.offering().html_label()} -[" + \
            f"{self.activity(),self.guard().html_label()}]->" + \
            f" {self.next().html_label()}"
    
@dataclass
class Offer():
    """ A data container for an offer from a transition tree. """

    sigma:Trace
    activity:Set[str]

    def __hash__(self):
        return hash((self.sigma.__hash__(), tuple(self.activity)))

class TransitionTree():
    """
    A transition tree is a connected, rooted, edge-labeled and directioned graph
    without cycles. See 'When are two  workflows the same?' In: CATS. ACS (2005).
    """

    def __init__(self, 
                 vertices:Set[TransitionTreeVertex], 
                 root:TransitionTreeRoot,
                 flows:Set[TransitionTreeFlow]) -> None:
        # check types and input
        if not (isinstance(vertices, set)) or vertices == None:
            raise ValueError(f"Unexpected vertices, was expecting a set but got"+
                             f" {type(vertices)}.")
        elif not (isinstance(root, TransitionTreeRoot)) or root == None:
            raise ValueError(f"Unexpected root, was expecting a "+
                             f"`TransitionTreeRoot` but got {type(root)}.")
        elif not (isinstance(flows, set)) or flows == None:
            raise ValueError("Unexpected flows, was expecting a set but got "+
                             f"{type(flows)}")
        # create copies
        self._vertices = deepcopy(vertices)
        self._root = deepcopy(root) 
        self._flows = deepcopy(flows)
        # precompute
        self._attrs = set([ a for f in self._flows for a in f.attributes() ])
        self._pops = [ 
            f.population() 
            for f 
            in self._flows 
            if isinstance(f, TransitionTreePopulationFlow)
        ]
        self._guards = set([
            f.guard()
            for f 
            in self._flows
            if isinstance(f, TransitionTreeGuardFlow)
        ])

    # properties
    def vertices(self) -> Set[TransitionTreeVertex]:
        """
        returns the full set of vertices that are defined by this transition tree.
        """ 
        return deepcopy(self._vertices)
    
    def terminals(self) -> Set[TransitionTreeVertex]:
        """
        returns the set of vertices that are terminal in this tree.
        """
        return set(
            deepcopy(node) 
            for node 
            in self._vertices
            if node.terminal()
        )

    def root(self) -> TransitionTreeRoot:
        """
        returns the root vertex of this transition tree.
        """ 
        return deepcopy(self._root)

    def flows(self) -> Set[TransitionTreeFlow]:
        """
        returns the full set of flows that are defined by this transition tree.
        """
        return deepcopy(self._flows)

    def strict_vertices(self) -> Set[TransitionTreeVertex]:
        """
        returns the partial set of vertices that have at least two outgoing flows.
        """
        out = set()
        flows = self.flows()
        for vertex in self.vertices():
            if len(vertex.outgoing(flows)) > 1:
                out.add(vertex)
        return out

    def offers(self) -> Set[Offer]:
        """
        returns the full set of offers that could happen based on this transition tree.
        """ 
        out = set()
        flows = self.flows()
        for vertex in self.vertices():
            acts = set()
            for flow in vertex.outgoing(flows):
                acts.add(flow.activity())
            out.add(Offer(vertex.sigma_sequence(), set(acts)))
        return out

    def attributes(self) -> Set[str]:
        """
        returns a set of attribute names that are used within the information attached to flows.
        """ 
        return deepcopy(self._attrs)

    def guards(self) -> Set:
        """
        returns the set of guards used in flows, that imply a population of data 
        attribute mappings.
        """
        return deepcopy(self._guards)

    def choices(self) -> Set:
        """
        returns a strict set of offerings, where we only consider offers from 
        vertices with many outgoing flows.
        """
        out = set()
        flows = self.flows()
        for vertex in self.strict_vertices():
            acts = set()
            for flow in vertex.outgoing(flows):
                acts.add(flow.activity())
            out.add(Offer(vertex.sigma_sequence(), set(acts)))
        return out

    def generate_dot(self, filepath:str) -> None:
        "Generates a dot file (for graphviz) for the transition tree."
        # string formats for nodes and arcs
        FILE_HEADER = """digraph G {
    rankdir="LR";
    bgcolor="white";
    forcelabels="true";
    pack="true";
    pad="0.1";
    compound="true";
    node[fixedsize="true", width="0.75", height="0.75", fontname="serif", fontsize="12", fontcolor="white", fillcolor="cadetblue",style="filled",shape="circle", label=""];
    edge[fontname="serif", fontcolor="black", labelfloat="true", minlen="5", fontsize="10", labeldistance="4"];\n\n"""
        ROOT_VERT = """\t\t{id} [label={html}, shape="doublecircle", fillcolor="aquamarine4"];\n"""
        TERM_VERT = """\t\t{id} [label={html}, shape="tripleoctagon", fillcolor="olivedrab4"];\n"""
        VERT = """\t\t{id} [label={html}];\n"""
        FLOW = """\t\t{offer} -> {next} [taillabel=<[{label},]>, headlabel=<<font color="darkgrey">f<sub>{id}</sub></font>>];\n"""
        GFLOW = """\t\t{offer} -> {next} [taillabel=<[{label},{guard}]>, headlabel=<<font color="darkgrey">f<sub>{id}</sub></font>>];\n"""
        FILE_FOOTER = """\n}"""
        # check if file exists
        dirname = path.dirname(filepath)
        if not path.exists(dirname):
            mkdir(dirname)
        with open(filepath, "w") as f:
            # add header
            f.write(FILE_HEADER)
            # add root
            root = self.root()
            f.write( 
                ROOT_VERT.format(id=root.id(), html=root.html_label())
            )
            # walk from root, adding vertices and flows
            flows = root.outgoing(self.flows())
            seen_flows = dict()
            seen_verts = set()
            currnt_flow = 1
            while len(flows) > 0:
                verts = [ flow.next() for flow in flows]
                for flow in flows:
                    if flow not in seen_flows.keys():
                        seen_flows[flow] = currnt_flow
                        currnt_flow += 1
                for v in verts:
                    if v.terminal():
                        f.write( 
                            TERM_VERT.format(id=v.id(), html=v.html_label())
                        )
                    else:
                        f.write( 
                            VERT.format(id=v.id(), html=v.html_label())
                        )
                for flow in flows:
                    if isinstance(flow, TransitionTreeGuardFlow):
                        f.write( 
                            GFLOW.format(
                                offer=flow.offering().id(),
                                next=flow.next().id(),
                                label=flow.activity(),
                                id=seen_flows[flow],
                                guard=flow.guard().html_label()
                            )
                        )
                    else:
                        f.write( 
                            FLOW.format(
                                offer=flow.offering().id(),
                                next=flow.next().id(),
                                label=flow.activity(),
                                id=seen_flows[flow]
                            )
                        )
                flows = []
                for v in verts:
                    if (v not in seen_verts):
                        flows += v.outgoing(self.flows())
                        seen_verts.add(v)
            # close and write footer
            f.write(FILE_FOOTER)
            
    # comparision functions

def apply_flow_reduction(tree:TransitionTree):
    """
    Ensures that only one flow exists between nodes in a tree.
    Returns a new tree with the same nodes but an alternative set of flows.
    """
    info("applying flow reduction on playout tree.")
    new_flows = set()
    for node in InfoIteratorProcessor("checking nodes", tree.vertices()):
        outgoing = node.outgoing(tree.flows())
        next_nodes = set( out.next() for out in outgoing )
        for next in next_nodes:
            act = next.sigma_sequence()
            act = act[len(act)-1]
            old_flows = set( out for out in outgoing if out.next() == next)
            if (len(old_flows) > 1):
                new_flows.add( 
                    TransitionTreeGuardFlow(
                        node,
                        act,
                        next,
                        TransitionTreeJoin( 
                            old_flows
                        )
                    )
                )
            else:
                new_flows = new_flows.union(old_flows)
    info("flow reduction complete.")
    return TransitionTree(
        tree.vertices(),
        tree.root(),
        new_flows
    )


def convert_playout_to_tree(playout_log:ComplexEventLog, k:int , 
                            freduce:bool=True) -> TransitionTree:
    """
    Converts a playout log to a transition tree, using a flow reduction step
    which may combine flows using a disjunction of guards.

    Parameters
    ----------
    `playout_log`: a log of play-out traces, see 
    pmkoalas.conformance.tokenreplay\n
    `k`: the length of the longest observed trace, used to limit tree depth.\n
    `freduce`: should flow reduction step be applied to ensure that only one
    flow exists between nodes.
    """
    # import here to avoid cirular dependenies
    from pmkoalas.conformance.tokenreplay import PlayoutEnd
    # construct root
    root = TransitionPlayoutRoot()
    # holder for ids
    class Idder():
        
        def __init__(self) -> None:
            self._counter = 1
        
        def __call__(self, *args: Any, **kwds: Any) -> Any:
            self._counter += 1
            return self._counter
    idder = Idder()
    # constuct nodes
    nodes = set()
    pbar = InfoQueueProcessor("built nodes of tree", 
                              starting_size=len(playout_log)
    )
    for _,instances in playout_log:
        for trace in instances:
            for i in range(0, min([k+1, len(trace)])):
                nodes.add(trace.acut(i+1))
            pbar.update()
    map_nodes = dict(
        (node, TransitionPlayoutVertex(idder(), node))
        for node
        in nodes
    )
    map_nodes[Trace([])] = root
    nodes.add(root)
    # construct flows
    flows = set()
    pbar = InfoQueueProcessor("built flows of tree",
                              starting_size=len(playout_log)
    )
    for _,instances in playout_log:
        for trace in instances:
            for i in range(0,min([k+1, len(trace)])):
                flows.add(
                    TransitionTreeGuardFlow(
                        map_nodes[trace.acut(i)],
                        trace.act(i),
                        map_nodes[trace.acut(i+1)],
                        trace.guard(i)
                    )
                )    
            pbar.update()
    # set final nodes
    pbar = InfoQueueProcessor("check for terminals",
                              starting_size=len(playout_log)
    )
    for _,instances in playout_log:
        for trace in instances:
            i = min([k+1, len(trace)-1])
            if isinstance(trace[i],  PlayoutEnd):
                map_nodes[trace.acut(i)].set_as_terminal()
            pbar.update()
    # construct tree
    tree = TransitionTree(
        set(list(map_nodes.values())),
        root,
        flows 
    )
    if (freduce):
        tree = apply_flow_reduction(tree)
    info("constructed tree.")
    return tree


def construct_from_model(model:object, longest_playout:int, freduce:bool=True)\
    -> TransitionTree:
    """ 
    Constructs a transition tree from an executable model.
    Currently only supports petri nets.

    Parameters
    ----------
    `model`: the model which will be used to generate the tree. 
    `longest_playout`: the length of the longest observed trace, used to limit tree depth.\n
    `freduce`: should flow reduction step be applied to ensure that only one
    flow exists between nodes.
    """
    # import here to avoid cirular dependenies
    from pmkoalas.models.petrinets.pn import AcceptingPetriNet
    from pmkoalas.conformance.tokenreplay import construct_playout_log
    # the work
    if (isinstance(model, AcceptingPetriNet)):
        playout_log = construct_playout_log(
            model, longest_playout,
            model.initial_marking,
            list(model.final_markings)[0]
        )
        return convert_playout_to_tree(playout_log, longest_playout, 
                                       freduce=freduce)
    else:
        raise ValueError("No known execution playout technique for model of" +\
                         f" type :: {type(model)}")


def construct_from_log(log:Union['EventLog','ComplexEventLog']) -> TransitionTree:
    """ Constructs a transition tree from an event log. """
    if (isinstance(log, EventLog)):
        return construct_from_simple_log(log)
    elif (isinstance(log, ComplexEventLog)):
        return construct_from_complex_log(log)
    else:
        raise ValueError("Was expecting a simple or complex event log, but was "+
                         f"given {type(log)}.")

def construct_from_simple_log(log:'EventLog') -> TransitionTree:
    """ Constructs a transition tree using a simple event log. """ 
    vroot = TransitionTreeRoot()
    vid = 1
    verts = dict( [(Trace([]), vroot)])
    flows = set()
    for trace,freq in log:
        pseq = []
        for act in trace:
            offering = verts[Trace(pseq)]
            if Trace(pseq + [act]) in verts.keys():
                next = verts[Trace(pseq + [act])]
            else:
                vid += 1
                last = False
                next = TransitionTreeVertex(
                    vid,
                    Trace(pseq + [act]),
                    end=last 
                )
                verts[Trace(pseq + [act])] = next
            flow = TransitionTreePopulationFlow( 
                offering,
                act,
                next,
                [ComplexEvent(act, dict())] * freq
            )
            flows.add(flow)
            pseq.append(act)
        # set the last vert as terminal
        offering = verts[Trace(pseq)]
        offering.set_as_terminal()
    # handle back
    verts = set([ v for v in verts.values()])
    return TransitionTree(verts, vroot, flows)

def construct_from_complex_log(log:'ComplexEventLog') -> TransitionTree:
    """ Constructs a transition tree using a complex event log. """
    # intial variables for tree
    vroot = TransitionTreeRoot()
    vid = 1
    verts = dict( [(Trace([]), vroot)])
    flows = set()
    # helper func to handle concat assignments
    def map_join(c:dict, n:dict) -> dict:
        out = dict()
        out.update(c)
        out.update(n)
        return out
    # finding all vertices and flows
    for trace,instances in log:
        pseq = []
        for id in range(0,len(trace)):
            act = trace[id]
            offering = verts[Trace(pseq)]
            if Trace(pseq + [act]) in verts.keys():
                next = verts[Trace(pseq + [act])]
            else:
                vid += 1
                last = False
                next = TransitionTreeVertex(
                    vid,
                    Trace(pseq + [act]),
                    end=last 
                )
                verts[Trace(pseq + [act])] = next
            flow = TransitionTreePopulationFlow( 
                offering,
                act,
                next,
                list()
            )
            # build population from instances
            assignments = [ 
                reduce(map_join,([ dict() ] + [ i[k].data() ]))
                for i
                in instances
                for k 
                in range(0, id-1)
            ]
            pops = [ ComplexEvent(act, assign) for assign in assignments ] 
            # check for existing population and update population
            if flow in flows:
                old_flow = [ f for f in flows if f == flow][0]
                pops = pops + old_flow.population()
                flows.remove(flow)
            flow = TransitionTreePopulationFlow( 
                offering,
                act,
                next,
                pops
            )
            flows.add(flow)
            pseq.append(act)
        # set last offering as terminal halt offering
        offering = verts[Trace(pseq)]
        offering.set_as_terminal()
    # handle back
    verts = set([ v for v in verts.values()])
    return TransitionTree(verts, vroot, flows)


"""
This modules provides ways to perfom data-aware conformance checking as
proposed by A. Banham, A. H. M. T. Hofstede, S. J. J. Leemans, F. Mannhardt,
R. Andrews and M. T. Wynn, "Comparing Conformance Checking for Decision 
Mining: An Axiomatic Approach," in IEEE Access, vol. 12, pp. 60276-60298, 
2024, doi: 10.1109/ACCESS.2024.3391234. 

As well as Determinism as proposed by Banham, Adam, Leemans, Sander, Wynn, 
Moe Thandar, & Andrews, Robert (2022) xPM: A Framework for Process Mining 
with Exogenous Data.

These conformance checking methods are related to the Petri net with data
process modelling formalism.

Implemented techniques include:
    - guard-recall 
    - guard-precision
    - determinism
"""
from pmkoalas.complex import ComplexEventLog
from pmkoalas.models.petrinets.dpn import PetriNetWithData
from pmkoalas.models.petrinets.dpn import GuardedTransition
from pmkoalas.models.petrinets.guards import GuardOutcomes
from pmkoalas.models.transitiontree import TransitionTreeGuardFlow
from pmkoalas.models.transitiontree import TransitionTreeGuardFlow
from pmkoalas.models.transitiontree import TransitionTree
from pmkoalas.models.transitiontree import construct_from_model
from pmkoalas.conformance.matching import ManyMatching, find_all_paths
from pmkoalas.conformance.matching import construct_many_matching
from pmkoalas.conformance.matching import _computation_many_matching
from pmkoalas.conformance.matching import ExpontentialPathWeighter
from pmkoalas._logging import info, enable_logging, debug
from pmkoalas._logging import InfoIteratorProcessor

from typing import Set
from joblib import Parallel, delayed

def compute_directed_bookkeeping(flow:TransitionTreeGuardFlow, 
    matcher:ManyMatching, log:ComplexEventLog, outcome=GuardOutcomes.TRUE) \
    -> float:
    """
    Computes the many directed bookkeeping on any flow for any given outcome.
    """
    bookkeeping = 0
    for trace, instances in log:
        weighter = ExpontentialPathWeighter(matcher[trace])
        paths = matcher[trace]
        for path in paths:
            for step,i in zip(path, range(1, len(path)+1)):
                if step == flow: 
                    i_w = 0
                    for instance in instances:
                        irveson = flow.guard().check(
                            instance.get_state_as_of(i-1)
                        ) == outcome
                        i_w += weighter(path, trace) if irveson else 0
                    bookkeeping += i_w
                    break
    return bookkeeping

def compute_general_bookkeeping(flow:TransitionTreeGuardFlow, 
    matcher:ManyMatching, log:ComplexEventLog, outcome=GuardOutcomes.TRUE) \
    -> float:
    """
    Computes the many general bookkeeping on any flow for any given outcome.
    """
    bookkeeping = 0
    for trace, instances in log:
        weighter = ExpontentialPathWeighter(matcher[trace])
        paths = matcher[trace]
        for path in paths:
            for step,i in zip(path, range(1, len(path)+1)):
                if step.offering() == flow.offering(): 
                    i_w = 0
                    for instance in instances:
                        irveson = flow.guard().check(
                            instance.get_state_as_of(i-1)
                        ) == outcome
                        i_w += weighter(path, trace) if irveson else 0
                    bookkeeping += i_w
                    break
    return bookkeeping

def compute_total_weight(flow:TransitionTreeGuardFlow, 
    matcher:ManyMatching, log:ComplexEventLog) -> float:
    """
    Computes the total weight on a flow, regradless of outcome.
    """
    weight = 0.0
    for trace, instances in log:
        paths = matcher[trace]
        for path in paths:
            for step in path:
                if step == flow: 
                    weight += (1/len(paths)) * len(instances)
                    break
    return weight

def _computation_guard_recall(tree:TransitionTree, matching:ManyMatching, 
    log:ComplexEventLog) -> float:
    """
    The actual computation of guard-recall using the structures.
    """
    flow_weight = set( 
        (flow, compute_total_weight(flow, matching, log))
        for flow
        in InfoIteratorProcessor("computing total weight", tree.flows(), 
                                 stack=4)
    )
    inner_sum = 0.0
    t_w = 0.0
    for flow,total_weight in InfoIteratorProcessor("computing flow weight", 
                                                   flow_weight):
        inner_sum += \
            compute_directed_bookkeeping(flow, matching, log)
        t_w += total_weight
    info(f"computed weights :: flow - {inner_sum:.2f} total - {t_w:.2f}")
    return (inner_sum/t_w)

def _optimised_guard_recall(log:ComplexEventLog, tree:TransitionTree,
        precomputed_matching:ManyMatching=None) -> float:
    """
    The computation of guard recall, whereby we only loop over the log once.
    """
    total_weight = 0
    flow_weight = 0
    # partial worker 
    pool = Parallel(n_jobs=-2)
    def partial(trace, instances, path, inst_w) -> float:
        flow_weight = 0
        total_weight = 0
        instance_weight = inst_w(path, trace)
        for step,i in zip(path, range(1, len(path)+1)):
                    if isinstance(step, TransitionTreeGuardFlow):
                        for instance in instances:
                            irveson = step.guard().check(
                                    instance.get_state_as_of(i-1)
                                ) == GuardOutcomes.TRUE
                            flow_weight += instance_weight * int(irveson)
        total_weight += inst_w.share * len(trace) * len(instances)
        return flow_weight, total_weight
    # worker for inputs
    def params(trace, instances, tree, allcads):
        if precomputed_matching != None:
            candidates = precomputed_matching[trace]
        else:
            _,candidates = _computation_many_matching(trace, tree, allcads)
        return trace, instances, candidates
    # computation of work
    max_k = max([len(trace) for trace,_ in log])
    allcads = find_all_paths(tree, max_k)
    info(f"size of all candidates given {max_k=} :: {len(allcads)}")
    inputs = list(pool( 
        delayed(params)
        (trace, instances, tree,
         set([ cad for cad in allcads if len(cad) <= len(trace)]))
        for trace,instances 
        in log 
    ))
    inputs = [
        (trace, instances, path, ExpontentialPathWeighter(matching))
        for trace, instances, matching
        in inputs 
        for path 
        in matching
    ]
    weights = pool( 
        delayed(partial)
        (traces, instances, path, inst_w)
        for traces, instances, path, inst_w 
        in InfoIteratorProcessor("processing variant-paths for guard-recall", 
            inputs,
            stack=8
        )
    )
    info("completed processing.")
    info(f"computed weights across flows (flow,total) :: {weights}")
    flow_weight += sum(
        f 
        for f,t 
        in weights
    )
    total_weight += sum( 
        t 
        for f,t 
        in weights
    )
    recall = flow_weight / total_weight
    info(f"computed guard recall of {recall:.3f}")
    return recall

@enable_logging
def compute_guard_recall(log:ComplexEventLog, model:PetriNetWithData, 
    optimised:bool=True, precomputed_matching:ManyMatching=None) -> float: 
    """
    Quanitfies the quality between an event log and a data-aware process model 
    using guard-recall, where a data-aware process model is a Petri net with 
    data.
    """
    # find longest observed trace
    long = None
    for trace,_ in log:
        if long == None:
            long = trace 
        elif len(trace) > len(long):
            long = trace 
    long = len(long)
    # prepare model
    tree = construct_from_model(model, longest_playout=long)
    if (optimised):
        return _optimised_guard_recall(log, tree, 
                precomputed_matching=precomputed_matching)
    else:
        info("creating matchings between all traces")
        if precomputed_matching == None:
            matching = construct_many_matching(log, tree)
        else:
            matching = precomputed_matching
        # compute the measure 
        return _computation_guard_recall(tree, matching, log)

def _computation_guard_precision(tree:TransitionTree, matching:ManyMatching,
    log:ComplexEventLog) -> float:
    """
    The actual computation of guard-precision using the structures.
    """
    def equal_share_bookkeeping(flow, mathching, log):
        bookkeeping = 0
        for trace, instances in log:
            weighter = ExpontentialPathWeighter(mathching[trace])
            paths = mathching[trace]
            for path in paths:
                for step,i in zip(path, range(1, len(path)+1)):
                    if step.offering() == flow.offering(): 
                        i_w = 0
                        for instance in instances:
                            irveson = flow.guard().check(
                                instance.get_state_as_of(i-1)
                            ) == GuardOutcomes.TRUE
                            i_w += weighter.share if irveson else 0
                        bookkeeping += i_w
                        break
        return bookkeeping
    upper_sum = 0.0
    lower_sum = 0.0
    for flow in InfoIteratorProcessor(
        "processing flows",
        tree.flows()):
        debug(f"processing flow {flow} :: {upper_sum=}/{lower_sum=}")
        upper_sum += compute_directed_bookkeeping(flow, matching, log)
        lower_sum += equal_share_bookkeeping(flow, matching, log)
        debug(f"processed flow {flow} :: {upper_sum=}/{lower_sum=}")
    prec = (1 + upper_sum) / (1 + lower_sum) 
    info(f"computed guard precision : {prec:.3f}")
    return prec

def _optimised_guard_precision(log:ComplexEventLog, tree:TransitionTree):
    """
    The computation of guard recall, whereby we only loop over the log once.
    """

    # partial worker 
    pool = Parallel(n_jobs=-2)
    def partial(tree, trace, instances, path, inst_w) -> float:
        upper_sum = 0.0
        lower_sum = 0.0
        instance_weight = inst_w(path, trace)
        for step,i in zip(path, range(1, len(path)+1)):
            if isinstance(step, TransitionTreeGuardFlow):
                other_flows = step.offering().outgoing(tree.flows())
                other_flows = other_flows.difference(set([step]))
                for instance in instances:
                    irveson = step.guard().check(
                            instance.get_state_as_of(i-1)
                        ) == GuardOutcomes.TRUE
                    upper_sum += instance_weight * int(irveson)
                    lower_sum += inst_w.share * int(irveson)
                    for lflow in other_flows:
                        irveson = lflow.guard().check(
                            instance.get_state_as_of(i-1)
                        ) == GuardOutcomes.TRUE
                        lower_sum += inst_w.share  * int(irveson)
        return upper_sum, lower_sum
    # worker for inputs
    info("preparing work")
    max_length = max([ len(trace) for trace,_ in log])
    allcads = find_all_paths(tree, max_length)
    def params(trace, instances, tree, allcads):
        _,candidates = _computation_many_matching(trace, tree, allcads)
        return tree, trace, instances, candidates
    # computation of work
    inputs = list(pool( 
        delayed(params)
        (trace, instances, tree, 
         set([ cad for cad in allcads if len(cad) <= len(trace)])
        )
        for trace,instances 
        in log 
    ))
    inputs = [
        (tree, trace, instances, path, ExpontentialPathWeighter(matching))
        for tree, trace, instances, matching
        in inputs 
        for path 
        in matching
    ]
    weights = pool( 
        delayed(partial)
        (tree, traces, instances, path, inst_w)
        for tree, traces, instances, path, inst_w 
        in InfoIteratorProcessor("processing variant-paths for guard-precision", 
            inputs,
            stack=8
        )
    )
    info("completed processing.")
    info(f"computed weights across flows (upper,lower) :: {weights}")
    upper_sum = sum(
        up 
        for up,low 
        in weights
    )
    lower_sum = sum( 
        low 
        for up,low 
        in weights
    )
    prec = (1 + upper_sum) / (1 + lower_sum)
    info(f"computed guard recall of {prec:.3f}")
    return prec

def compute_guard_precision(log:ComplexEventLog, model:PetriNetWithData,
    optimised:bool=True ) -> float:
    """
    Quanitfies the quality between an event log and a data-aware process model 
    using guard-precision, where a data-aware process model is a Petri net with 
    data.
    """
    # find longest observed trace
    long = None
    for trace,_ in log:
        if long == None:
            long = trace 
        elif len(trace) > len(long):
            long = trace 
    long = len(long)
    # prepare model
    tree = construct_from_model(model, longest_playout=long)
    if optimised:
        return _optimised_guard_precision(log, tree)
    else:
        matching = construct_many_matching(log, tree)
        # compute measure
        return _computation_guard_precision(tree, matching, log)
    
@enable_logging
def compute_determinism(model:PetriNetWithData) -> float:
    """
    Computes the determinism of a Petri net with Data. Based on the defintion
    in Banham, Adam, Leemans, Sander, Wynn, Moe Thandar, & Andrews, Robert 
    (2022) xPM: A Framework for Process Mining with Exogenous Data. 
    Determinism is useful for showing if how many transitions have been
    annotated with a guard from a decision mining technique.

    Determinism expresses the decision points in the model that may be 
    deterministic. That is, the fraction of transitions in the model, that:
        - are in the postset of a place with with two or more outgoing arcs
          or are decision points, and
        - that have an associated (non-trivial) precondition (defined and 
          not equal to a literal truth).
    
    Returns a measure between 0 and 1.
    A value of 1 for determinism implies that all transitions that are involved
    in choices in the model have discovered preconditions, while a value of 0 
    indicates that no transition that is involved in a choice has a discovered
    precondition.
    """
    # find all transitions that are in the postsets of places 
    # with two or more transitions
    dtrans:Set[GuardedTransition] = set()
    for place in model.places:
        outgoing_nodes = set([
            arc.to_node
            for arc
            in model.arcs
            if arc.from_node == place
        ])
        if (len(outgoing_nodes) >= 2):
            dtrans = dtrans.union(outgoing_nodes)
    info(f"number of decision transitions: {len(dtrans)}")
    # find the subset of transitions with an associated 
    atrans:Set[GuardedTransition] = set()
    for trans in dtrans:
        if not trans.guard.trivial():
            atrans.add(trans)
    info(f"number of associated transitions: {len(atrans)}")
    # return measurement
    if len(dtrans) > 0:
        return (len(atrans) * 1.0) / (len(dtrans) * 1.0)
    else:
        return 1



""""
This module contains the abstractions and structures for a data Petri net.

"""
from pmkoalas.models.petrinets.pn import LabelledPetriNet, Place
from pmkoalas.models.petrinets.pn import Transition, Arc
from pmkoalas.models.petrinets.pn import AcceptingPetriNet, PetriNetMarking
from pmkoalas.models.petrinets.pn import PetriNetSemantics
from pmkoalas.models.petrinets.pn import ExecutablePetriNet
from pmkoalas.models.petrinets.guards import Guard

from copy import deepcopy
from typing import Iterable, FrozenSet, Set

class GuardedTransition(Transition):
    """
    An abstraction for a transition with a guard.
    """

    def __init__(self, 
                 name: str, 
                 guard: Guard,
                 tid: str = None, 
                 silent: bool = False):
        super().__init__(name, tid, silent)
        self._guard = deepcopy(guard)

    @property
    def guard(self) -> Guard:
        return self._guard
    
    def __str__(self) -> str:
        return f"[{self.name} {self.guard}]"
    
    def __hash__(self) -> int:
        return hash((super().__hash__(), self.guard.__hash__()))
    
    def __eq__(self, other) -> bool:
        if super().__eq__(other):
            return self.guard == other.guard
        return False
    
    def __repr__(self) -> str:
        return f'GuardedTransition("{self.name}",guard={self.guard.__repr__()}'\
               + f',tid="{self.tid}",silent={self.silent})'
    
class PetriNetWithData(LabelledPetriNet):
    """
    An abstraction for extending a Petri net to one with data. This abstraction
    only includes guards.
    """

    def __init__(self, places: Iterable[Place], 
                 transitions: Iterable[GuardedTransition], 
                 arcs: Iterable[Arc], 
                 name: str = 'Petri net with Data'):
        super().__init__(places, transitions, arcs, name)

    @property
    def transitions(self) -> FrozenSet[GuardedTransition]:
        return deepcopy(self._transitions)
    
    def __repr__(self) -> str:
        repr = "PetriNetWithData(\n"
        # add places
        repr += "\tplaces=[\n"
        for p in self.places:
            repr += f"\t\t{p.__repr__()},\n"
        repr += "\t],\n"
        # add transitions
        repr += "\ttransitions=[\n"
        for t in self.transitions:
            repr += f"\t\t{t.__repr__()},\n"
        repr += "\t],\n"
        # add arcs
        repr += "\tarcs=[\n"
        for a in self.arcs:
            repr += f"\t\t{a.__repr__()},\n"
        repr += "\t],\n"
        # add name
        repr += f"\tname='{self.name}'\n"
        #close param
        repr += ")"
        return repr
    
class AcceptingDataPetriNet(AcceptingPetriNet):
    """
    Further structure for a data Petri net that contains its initial marking
    and possible final markings.
    """
    def __init__(self, net:PetriNetWithData, inital:PetriNetMarking, 
                 finals:Set[PetriNetMarking]):
        super().__init__(net, inital, finals)

    @property
    def net(self) -> PetriNetWithData:
        return super().net

class DataPetriNetSemantics(PetriNetSemantics):
    """
    An abstraction for the semantics of a data Petri net.
    """
    def __init__(self, net: PetriNetWithData, curr: PetriNetMarking):
        super().__init__(net, curr)

    def fireable(self) -> Set[GuardedTransition]:
        return super().fireable()
    
    def peek(self, firing: GuardedTransition) -> 'DataPetriNetSemantics':
        return super().peek(firing)
    
    def fire(self, firing: GuardedTransition) -> 'DataPetriNetSemantics':
        return super().fire(firing)

class ExecutableDataPetriNet(ExecutablePetriNet):
    """
    An abstraction for a data Petri net that can be executed.
    """
    def __init__(self, anet:AcceptingDataPetriNet, sem:DataPetriNetSemantics):
        super().__init__(anet, sem)

    @property
    def semantics(self) -> DataPetriNetSemantics:
        return super().semantics
    
    @property
    def anet(self) -> AcceptingDataPetriNet:
        return super().anet

def get_execution_semantics(anet:AcceptingDataPetriNet) -> ExecutableDataPetriNet:
    """
    Constructs an executable data Petri net from the given data Petri net and
    initial marking.
    """
    sem = DataPetriNetSemantics(anet, anet.initial_marking)
    return ExecutableDataPetriNet(anet, sem)


"""
This module contains helpers to create pretty dot (graphviz) digraphs 
for petri-net-type structures.

"""
from pmkoalas.models.petrinets.pn import LabelledPetriNet

class PetriNetDOTFormatter:
    """
    This class creates a dot (graphviz) structure for displaying Petri nets. 
    Used as internal machinery for exporting to a dot file.
    """
    def __init__(self,pn:LabelledPetriNet,font:str='SimSun'):
        self._pn = pn
        self._font = font
        self._nodemap = {}
        self._default_height = 0.2

    def transform_transition(self,tran,ti) -> str:
        fstr = 'n{} [shape="box",margin="0, 0.1",label="{} {}",style="filled"];\n'
        tl = tran.name if tran.name and tran.name != 'tau' else '&tau;'
        fstr = f'n{str(ti)} [shape="box",margin="0, 0.1",'
        # fstr += f'label="{tran.weight}", style="filled",'
        height = self._default_height
        fstr += f'height="{height}", width="{height}"'
        fstr += '];\n'
        return fstr

    def transform_place(self,place,pi) -> str:
        fstr = '{} [shape="circle",label="{}"];\n'
        return fstr.format('n' + str(pi),place.name)

    def transform_arc(self,arc) -> str:
        from_node = self._nodemap[arc.from_node.nodeId]
        to_node = self._nodemap[arc.to_node.nodeId]
        return f'n{from_node}->n{to_node}\n'

    def transform_net(self) -> str:
        dotstr = ""
        dotstr += 'digraph G{\n'
        dotstr += f'ranksep=".3"; fontsize="14"; remincross=true; margin="0.0,0.0"; fontname="{self._font}";rankdir="LR";charset=utf8;\n'
        dotstr += 'edge [arrowsize="0.5"];\n'
        dotstr += f'node [height="{self._default_height}",width="{self._default_height}",fontname="{self._font}"'
        dotstr += ',fontsize="14"];\n'
        dotstr += 'ratio=0.4;\n'
        ni = 1
        for pl in self._pn.places:
            ni += 1
            self._nodemap[pl.nodeId] = ni
            dotstr += self.transform_place(pl,ni)
        for tr in self._pn.transitions:
            ni += 1
            self._nodemap[tr.nodeId] = ni
            dotstr += self.transform_transition(tr,ni)
        for ar in self._pn.arcs:
            dotstr += self.transform_arc(ar)
        dotstr += '}\n'
        return dotstr

def convert_net_to_dot(net:LabelledPetriNet) -> str:
    return PetriNetDOTFormatter(net).transform_net()


"""
This module handles all machinery need to parse, check, and export
expressions on models, i.e. guards associated with transitions in a Petri net
with data.
"""
from pmkoalas._logging import debug
from pmkoalas.complex import ComplexTrace

from pyparsing import (
    Word,
    nums,
    alphas,
    printables,
    Combine,
    oneOf,
    opAssoc,
    infixNotation,
    Literal,
    ParserElement
)
ParserElement.enablePackrat()

from typing import Any,Set,Dict, Union
from copy import deepcopy,copy
from enum import Enum

class EvalLiteral:
    "Class to evaluate a parsed literal"

    def __init__(self):
        self.value = None

    def eval(self):
        return eval( self.value.replace('&#34;',"'"))

    def __call__(self, tokens) -> Any:
        self.value = tokens[0]
        return copy(self)
    
class EvalVariable:
    "Class to evaluate a parsed variable"

    def __init__(self, state_space):
        self.value = None
        self._observed_variables = set()
        self._set_state_space(state_space)

    def _set_state_space(self, space):
        self._state_space = space 

    def eval(self, dryrun=False):
        if (dryrun):
            self._observed_variables.add(self.value)
            return
        if self.value in self._state_space:
            return self._state_space[self.value]
        raise ValueError(f"Undefined variable :: {self.value}")

    def seen_variables(self) -> Set:
        return self._observed_variables
    
    def __call__(self, tokens) -> Any:
        self.value = tokens[0]
        self.eval(dryrun=True)
        return copy(self)
    
class EvalConstant:
    "Class to evaluate a parsed numeric constant"

    def __init__(self):
        self.value = None

    def eval(self):
        if self.value == 'true':
            return True 
        elif self.value == 'false':
            return False
        return float(self.value)

    def __call__(self, tokens) -> Any:
        self.value = tokens[0]
        return copy(self)


class EvalComparisonOp:
    "Class to evaluate comparison expressions"
    opMap = {
        "&lt;" : lambda a, b: a < b,
        "&lt;=" : lambda a,b: a <= b,
        "&gt;" : lambda a, b: a > b,
        "&gt;=" : lambda a,b: a >= b,
        "==": lambda a, b: a == b,
        "!=": lambda a, b: a != b,
        '&amp;&amp;': lambda a,b: a and b,
        "<" : lambda a, b: a < b,
        "<=" : lambda a,b: a <= b,
        ">" : lambda a, b: a > b,
        ">=" : lambda a,b: a >= b,
        '&&': lambda a,b: a and b,
        '||': lambda a,b: a or b
    }

    def __init__(self):
        self.value = None

    def eval(self):
        val1 = self.value[0].eval()
        debug(f"starting with left of :: {val1}")
        for op, val in self.operatorOperands(self.value[1:]):
            fn = EvalComparisonOp.opMap[op]
            val2 = val.eval()
            debug(f"operating on :: {val1} {op} {val2}")
            if not fn(val1, val2):
                break
            val1 = val2
        else:
            debug("resulting in true")
            return True
        debug("resulting in false")
        return False

    def operatorOperands(self,tokenlist):
        "generator to extract operators and operands in pairs"
        it = iter(tokenlist)
        while 1:
            try:
                yield (next(it), next(it))
            except StopIteration:
                break

    def __call__(self, tokens) -> Any:
        self.value = tokens[0]
        return copy(self)

class ExpressionParser():
    "A parser for guards associated with Petri net with Data"

    def __init__(self, state_space, exp:str) -> None:
        self._space = state_space
        self._org_exp = exp
        self._create_instances()
        self._create_operand()
        self._expr_form()
        self._parse()

    def _create_instances(self) -> None:
        self._constant = EvalConstant()
        self._literal = EvalLiteral()
        self._variable = EvalVariable(self._space)
        self._comparison = EvalComparisonOp()

    def _create_operand(self) -> None:
        # operand for parser
        booleans = Literal("true") | Literal('false')
        integer = Word(nums)
        pos_real = Combine(Word(nums) + "." + Word(nums))
        neg_real = Combine("-"+Word(nums) + "." + Word(nums))
        values = booleans | pos_real | neg_real | integer 
        values.setParseAction(self._constant)

        non_op_chars = "".join([c for c in printables if c not in "<>=!&|'\""])

        literal = Combine('&#34;' + Word(alphas+nums+non_op_chars+" ") + '&#34;')
        literal = literal | Combine('"' + Word(alphas+nums+non_op_chars+" "+"/") + '"')
        literal.setParseAction(self._literal)

        variable = Combine(Word(alphas,exact=1) + Word(alphas + nums + "_:$"))
        variable.setParseAction(self._variable)

        self._operand = values | literal | variable

    def _expr_form(self) -> None:
        comparisonop = oneOf("&lt;= <= &gt;= >= &lt; < &gt; > == !=")
        self._expr = infixNotation(
            self._operand,
            [
                (comparisonop, 2, opAssoc.LEFT, self._comparison),
                (oneOf("&amp;&amp; && ||"), 2, opAssoc.LEFT,  self._comparison)
            ],
        )

    def _parse(self):
        try :
            self._result = self._expr.parseString(self._org_exp)[0]
        except Exception as e:
            print(f"Failed to parse the following string :: {self._org_exp}")
            raise e

    def get_observed_vars(self):
        return self._variable.seen_variables()

    def change_state_space(self, state_space):
        self._space = state_space
        self._create_instances()
        self._create_operand()
        self._expr_form()
        self._parse()

    def result(self):
        return self._result.eval()
    
class GuardOutcomes(Enum):
    TRUE = True 
    FALSE = False
    UNDEF = "undefined"
    
class Expression():
    """
    A representation of reasoning extractable from a boolean logic expression.
    """

    def __init__(self, exp:str,) -> None:
        self._org_exp = exp 
        self._parser = ExpressionParser(dict(), exp)
        self._parsed_exp = None
        self._dom = self._find_dom(exp)

    def _find_dom(self, exp:str) -> Set[str]:
        parser = ExpressionParser(dict(), exp)
        return parser.get_observed_vars()
    
    def literal_truth(self) -> bool:
        """
        Evaluates if this expression is a literal truth, i.e. is always
        true and has no variables.
        """
        if len(self._dom) == 0:
            return self.evaluate({})
        else:
            return False
    
    def can_evaluate(self, data:Dict[str,object]) -> bool:
        """
        Checks if all variables are present for evaluation.
        """
        for req in self._dom:
            if req not in data:
                return False
        return True

    def evaluate(self, data:Dict[str,object]) -> GuardOutcomes:
        """
        Evalutes the expression using the given data.
        """
        if (not self.can_evaluate(data)):
            return GuardOutcomes.UNDEF
        try:
            ret = ExpressionParser(data, self._org_exp).result()
            if ret:
                return GuardOutcomes.TRUE
            else:
                return GuardOutcomes.FALSE
        except Exception as e:
            debug(f"Failed to evaluate :: {e}")
            return GuardOutcomes.UNDEF
        
    def variables(self) -> Set[str]:
        """
        Returns the set of variables used in the expression.
        """
        return self._dom

    def __str__(self) -> str:
        return self._org_exp
    
    def __eq__(self, other: object) -> bool:
        if (type(self) == type(other)):
            return hash(self) == hash(str)
        return False
    
    def __hash__(self) -> int:
        return hash(str(self))
    
    def __repr__(self) -> str:
        return f"Expression('{self._org_exp}')"

class Guard():
    """
    An abstraction for a guard.
    """

    def __init__(self, exp:Union[str,Expression]) -> None:
        # convert string into expression if needed
        if isinstance(exp, str):
            exp = Expression(exp)
        self._exp = exp

    def variables(self) -> Set[str]:
        """
        Returns the set of variables used in the guard.
        """
        return self._exp.variables()

    def evaluate(self, trace:ComplexTrace, i:int) -> GuardOutcomes:
        """
        Evalutes the guard in the context of the given trace, before the i-th 
        event.
        """
        state = trace.get_state_as_of(i)
        outcome = self._exp.evaluate(state)
        return outcome
    
    def evaluate_data(self, data:Dict[str,object]) -> GuardOutcomes:
        """
        Evaluates the guard in the context of the given data state.
        """
        return self._exp.evaluate(data)
    
    def trivial(self) -> bool:
        """
        Returns True if the guard is a literal truth, .e. is always
        true and has no variables.
        """
        return self._exp.literal_truth()

    def __str__(self) -> str:
        return str(self._exp)
    
    def __hash__(self) -> int:
        return self._exp.__hash__()
    
    def __eq__(self, other: object) -> bool:
        if(type(self) == type(other)):
            return hash(self) == hash(other)
        return False
    
    def __repr__(self) -> str:
        return f'Guard({self._exp.__repr__()})'


"""
This module contains information about the structure of complex
api for process data, being how to represent an event,
a trace and an event log in complex form, which considers the
data perspective.
"""
from __future__ import annotations # required for typing checks
from typing import Mapping, Iterable, Set, List, Tuple
from copy import deepcopy
from time import time

from pmkoalas._logging import info, debug, enable_logging
from pmkoalas.simple import Trace, EventLog

# only evaluate these classes if we are type checking/hinting 
# prevents cyclic imports
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pmkoalas.models.transitiontree import TransitionTree

class ComplexEvent():
    """
    A complex form of an event, an atomic change in the state
    of a process.
    """

    STR_FORMAT = "(e : {act}|{map})"

    def __init__(self, activity:str, data:Mapping[str,object]) -> None:
        self._act = activity
        self._map = deepcopy(data)

    def activity(self) -> str:
        """ the process activity denoted by this event """
        return self._act

    def data(self) -> Mapping[str,object]:
        """ returns a copy of the data attached at this event """
        return deepcopy(self._map)    

    # data model functions
    def __getitem__(self, key):
        if key in self._map.keys():
            return deepcopy(self._map[key])
        else:
            return "UNDEFINED"
        
    def _pretty_map(self) -> str:
        str_map = "{"
        for key,val in self.data().items():
            str_map += f" '{key}' : {val},"
        return str_map[:-1] + "}"

    def __str__(self) -> str:
        label = self.STR_FORMAT.format(act=self._act,map=self._pretty_map())
        return label

    def __repr__(self) -> str:
        repr = "ComplexEvent(\n\t"
        repr += f"'{self._act}'" + ", {\n"
        for key,val in self._map.items():
            repr += f"\t'{key}' : {val.__repr__()},\n"
        repr += "\t}\n)"
        return repr
    
    def __hash__(self) -> int:
        return hash(
            tuple(
                [self._act,]+
                [hash(tuple([key,val])) for key,val 
                 in self._map.items()]
                )
        )
    
    def __eq__(self, value):
        if (isinstance(value, ComplexEvent)):
            return self.activity() == value.activity() \
                   and self.data() == value.data()
    
class ComplexTrace():
    """
    A complex form of sequence of complex events, where a trace
    can have a invariant mapping attached for trace level attributes.
    """

    def __init__(self, events:Iterable[ComplexEvent], 
                 data: Mapping[str,object] = None) -> None:
        self._sequence = [ 
            deepcopy(event)
            for event
            in events
        ]
        self._len = len(self._sequence)
        if data == None:
            self._map = dict()
        elif isinstance(data,(dict)):
            self._map = deepcopy(data)
        else:
            raise ValueError(f"Given data is not a map/dict :: {type(data)}")
        self._hash = hash( 
            tuple(list(self._map.items()) + [ s.__hash__() for s in self._sequence])
        )
        self._acts = set([ s.activity() for s in self._sequence])

    # accessors
    def get_id(self) -> str:
        return "complex"
    
    def seen_activities(self) -> Set[str]:
        """ returns the activities seen in this trace """
        return deepcopy(self._acts)
    
    def get_state_as_of(self, i:int) -> Mapping[str,object]:
        """ returns the data state of the trace before the i-th event."""
        state = dict()
        curr = 0
        while i > 0:
            if (curr < len(self)):
                state.update(self[curr].data())
            i -= 1
        return state
    
    def data(self) -> Mapping[str,object]:
        """ returns the trace attributes """
        return deepcopy(self._map)
    
    def simplify(self) -> Trace:
        """ 
        returns a simplifed representation of this trace without data.
        """
        return Trace([ s.activity() for s in self])
    
    # data model functions
    def __str__(self) -> str:
        if (self._len < 1):
            return f"<>:{str(self._map)}"
        else:
            rep = "<"
            for event in self:
                rep = rep + f"{str(event)}, "
            rep = rep[:-2] + ">"
            return rep

    def __repr__(self) -> str:
        repr = "ComplexTrace(\n\t[\n" 
        # add events in sequence
        for ev in self:
            repr += "\t\t" + ev.__repr__() + ",\n"
        repr += "\t],\n"
        # add map
        repr += "\tdata= "+ self.data().__repr__() +"\n" 
        return repr + ")"
    
    def __iter__(self) -> Iterable[ComplexEvent]:
        for event in self._sequence:
            yield event 

    def __getitem__(self,no:int) -> ComplexEvent:
        if (isinstance(no,(int))):
            no = int(no)
            if (no < self._len and no >= 0):
                return self._sequence[no]
            else:
                raise ValueError("Invalid sequence reference, "+ 
                                 f"must be between 0 and {self._len}"+
                                 f" but was given {no}.")
        else:
            raise ValueError(
                "Invalid sequence reference, expecting a int"+
                f" but was given {type(no)}."      
            )

    def __eq__(self, __o: object) -> bool:
        if ( isinstance(__o, ComplexTrace )  ):
            if (self._map.__eq__(__o._map)): 
                if (self._len == __o._len):
                    for e,oe in zip(self, __o):
                        if (not e.__hash__() == oe.__hash__()):
                            return False 
                    return True
                else:
                    return False
            else:
                return False
        return False

    def __hash__(self) -> int:
        return self._hash

    def __len__(self) -> int:
        return self._len

DEFAULT_COMPLEX_LOG_NAME = "complex log"
class ComplexEventLog():
    """
    A complex collection of complex traces and can have an 
    invariant mapping. This collection can express several 
    languages.
    """

    @enable_logging
    def __init__(self, traces: Iterable[ComplexTrace],
                 data:Mapping[str,object] = None,
                 name:str = DEFAULT_COMPLEX_LOG_NAME) -> None:
        # middleman to multi set repr
        self._freqset = dict()
        self._instances = dict()
        self._len = 0
        self._variants = 0
        self._pop_size = 0
        self._acts = set([])
        self._start_acts = set([])
        self._end_acts = set([])
        self._traces = None
        self._map = deepcopy(data)
        info("Computing language...")
        start = time()
        for trace in traces:
            
            strace = trace.simplify()
            if (strace in self._instances):
                collector = self._instances[strace]
                self._pop_size += 1
                collector.append(trace)
            else:
                self._instances[strace] = list()
                self._instances[strace].append(trace)
                self._pop_size += 1
            trace = strace
            if (trace in self._freqset.keys()):
                self._freqset[trace] += 1
            else:
                self._acts = self._acts.union(
                    trace.seen_activities()
                )
                if (len(trace) > 0):
                    self._start_acts.add(trace[0])
                    self._end_acts.add(trace[-1])
                self._freqset[trace] = 1
                self._variants += 1
            self._len += 1
        self._traces = set([ t for t in self._freqset.keys() ])
        info(f"Computed language in {(time()-start)*1000:.0f}ms")
        self.name = name 

    @enable_logging
    def simplify(self) -> EventLog:
        """
        Simplifies this event log to only consider the process
        activities. Returns a new instance of simple event log.
        """
        simple_traces = []
        for trace, freq in self._freqset.items():
            simple_traces = simple_traces + ([ trace ] * freq)
        return EventLog(simple_traces, self.name)

    def seen_activities(self) -> Set[str]:
        "Get a language of process activities from this language."
        return deepcopy(self._acts)

    def seen_start_activities(self) -> Set[str]:
        "Get a set of start activities from this language."
        return deepcopy(self._start_acts)

    def seen_end_activities(self) -> Set[str]:
        "Get a set of end activities from this language."
        return deepcopy(self._end_acts)
    
    def simple_language(self) -> Set[Trace]:
        "Get a simplified trace language from this language."
        return set(list(self._freqset.keys()))
    
    def simple_stochastic_language(self) -> Mapping[Trace,float]:
        "Get a simplified stochastic language from this language."
        return self._freqset.copy()
    
    def get_instances(self) -> Mapping[Trace, List[ComplexTrace]]:
        """ 
        Get a map between seen simple traces and instances 
        of complex traces.
        """
        return deepcopy(self._instances)
    
    def seen_instances_for(self, trace:Trace) -> Set[ComplexTrace]:
        """
        Explores this collection for instances of the given 
        simplified trace, may return instances or an empty collection.
        """
        if (trace in self._traces):
            return deepcopy(self._instances[trace]) 
        else:
            return []
        
    def get_name(self) -> str:
        " returns the name of this collection."
        return self.name

    def get_nvariants(self) -> int:
        " returns the number of trace variants in this collection."
        return self._variants
    
    def get_population_size(self) -> int:
        " returns the number of traces instances in this collection."
        return self._pop_size
    
    def data(self) -> Mapping[str,object]:
        " returns a data mapping for this collection, i.e. log-level attributes."
        return self._map
    
    # data model functions
    def __len__(self) -> int:
        return self.get_population_size()
    
    def __iter__(self) -> Iterable[Tuple[Trace,List[ComplexTrace]]]:
        for strace, collector in self._instances.items():
            yield deepcopy(strace), deepcopy(collector)

    def __str__(self) -> str:
        if (self._variants < 1):
            return "[]"

        _str = "[\n"
        for strace,complexes in self:
            for complex in complexes:
                _str = _str + "\t" + str(complex)+":"+str(complex.data())+",\n"
        _str = _str[:-2] + "\n]"
        return _str
    
    def __repr__(self) -> str:
        repr = "ComplexEventLog(\n\t[\n"
        # add complex traces
        for _, complexes in self:
            for complex in complexes:
                repr = repr + "\t" + complex.__repr__() +",\n"
        repr = repr +"\t],\n"
        # add data for event log
        repr = repr + "\tdata="+self._map.__repr__()+"\n"
        return repr + ")"
    
    def __eq__(self, __o: object) -> bool:
        if (isinstance(__o, ComplexEventLog)):
            if (__o.get_nvariants() == self.get_nvariants()):
                if (__o.get_population_size() == self.get_population_size()):
                    return self.get_instances() == __o.get_instances()  
        elif (isinstance(__o, EventLog)):
            return self.simplify() == __o
        # fall through
        return False




"""
This model contains the base abstraction layers for exporting a net into
a pnml file.
"""
from pmkoalas.models.petrinets.pn import LabelledPetriNet
from pmkoalas.models.petrinets.dpn import PetriNetWithData
from pmkoalas.models.petrinets.pn import AcceptingPetriNet
from pmkoalas.models.petrinets.dpn import GuardedTransition
from pmkoalas.models.petrinets.wpn import WeightedTransition

from uuid import UUID
import xml.etree.ElementTree as ET
from typing import Union

ENCODING='unicode'
PNML_URL='http://www.pnml.org/version-2009/grammar/pnmlcoremodel'

def convert_net_to_xml(
        net:Union[LabelledPetriNet,AcceptingPetriNet],
        include_prom_bits:bool=True       
    ) -> ET.Element: 
    """
    Converts a given Petri net to an XML structure that conforms with the pnml
    schema.

    See: http://www.pnml.org/version-2009/grammar/pnmlcoremodel.rng
    """
    if (isinstance(net,AcceptingPetriNet)):
        anet = net
        net = net.net
    else:
        anet = net
    # handlers for UUID for prombits
    from random import Random
    rd = Random(1998)
    def getUUID(given:str) -> str:
        try:
            uuid_obj = UUID(given, version=4)
            if str(uuid_obj) == given:
                return given
            raise ValueError("given was not a valid UUID")
        except Exception:
            return str(UUID(int=rd.getrandbits(128), version=4))
    # the conversion
    root = ET.Element('pnml')
    net_node = ET.SubElement(root,'net', 
            attrib={'type': PNML_URL,
                    'id':net.name} )
    net_name = ET.SubElement(net_node, "name")
    net_text = ET.SubElement(net_name, "text")
    net_text.text = net.name
    page = ET.SubElement(net_node,'page', id="page1")
    nodes = dict()
    attributes_for_guards = set()
    for place in net.places:
         # work out the id for the transition
        if isinstance(place.pid, int):
            pid = f"{place.pid}"
        else:
            pid = place.pid
        nodes[place] = pid
        placeNode = ET.SubElement(page,'place', 
            attrib={'id': pid } )
        if place.name:
            name_node = ET.SubElement(placeNode,'name')
            text_node = ET.SubElement(name_node,'text')
            text_node.text = place.name
        if (include_prom_bits):
            ET.SubElement(
                    placeNode, 'toolspecific',
                    attrib={
                        'tool' : "ProM",
                        'version' : "6.4",
                        'localNodeID' : getUUID(place.pid)
                    }
                )
        if isinstance(anet,AcceptingPetriNet):
            if anet.initial_marking.contains(place):
                imarking = ET.SubElement(placeNode,'initialMarking')
                text_node = ET.SubElement(imarking,'text')
                text_node.text = str(anet.initial_marking[place])
            fmark = list(anet.final_markings)[0]
            if fmark.contains(place):
                fmarking = ET.SubElement(placeNode,'finalMarking')
                text_node = ET.SubElement(fmarking,'text')
                text_node.text = str(fmark[place])

    for tran in net.transitions:
        # work out the id for the transition
        if isinstance(tran.tid, int):
            tid = f"{tran.tid}"
        else:
            tid = tran.tid
        nodes[tran] = tid
        # the default attributes for transitions
        attribs = { 'id' : tid}
        # check for guards 
        if isinstance(tran, GuardedTransition):
            attribs['guard'] = str(tran.guard)
            attributes_for_guards.update(tran.guard.variables())
        # check for silent
        if tran.silent:
            attribs['invisible'] = "true"
        # make a transition
        tranNode = ET.SubElement(page,'transition', 
                        attrib=attribs )
        if tran.name:
            name_node = ET.SubElement(tranNode,'name')
            text_node = ET.SubElement(name_node,'text')
            text_node.text = tran.name
        # include additional info for prom
        if (include_prom_bits):
            prom_node = ET.SubElement(
                tranNode, 'toolspecific',
                attrib={
                    'tool' : "ProM",
                    'version' : "6.4",
                    'activity' : "$invisible$" if tran.silent else "",
                    'localNodeID' : getUUID(tran.tid)
                }
            )
        # include stochastic info
        if isinstance(tran,WeightedTransition):
            ET.SubElement(tranNode,'toolspecific',
                        attrib={ 'tool':'StochasticPetriNet',
                                'version':'0.2', 
                                'invisible': str(tran.silent),
                                'priority': '1',
                                'weight' : str(tran.weight),
                                'distributionType': 'IMMEDIATE'} 
            )
    arcid = 1
    for arc in net.arcs:
        aid = "arc-"+str(arcid)
        arcNode = ET.SubElement(page,'arc',
            attrib={'source': nodes[arc.from_node], 
                    'target': nodes[arc.to_node],
                    'id':  aid} )
        arcid += 1
        if (include_prom_bits):
            prom_node = ET.SubElement(
                arcNode, 'toolspecific',
                attrib={
                    'tool' : "ProM",
                    'version' : "6.4",
                    'localNodeID' : getUUID(aid)
                }
            )
            arctype = ET.SubElement(
                arcNode, "arctype"
            )
            ET.SubElement(
                arctype, "text"
            ).text = "normal"

    # check if dpn and if so add variables
    if (isinstance(net,PetriNetWithData)):
        vars = ET.SubElement(net_node, 'variables')
        for attr in attributes_for_guards:
            var = ET.SubElement(vars, 'variable', attrib={'type':'java.lang.Double'})
            name = ET.SubElement(var, 'name')
            name.text = attr
    
    return root


def convert_net_to_xmlstr(
        net:LabelledPetriNet,
        include_prom_bits:bool=False             
    ) -> str: 
    """
    Converts a given Petri net, to an XML structure, and then returns a string
    representation of the indented XML tree.
    """
    xml = convert_net_to_xml(net,include_prom_bits)  
    ET.indent( xml ) 
    return ET.tostring(xml,encoding=ENCODING)

def export_net_to_pnml(
        net:LabelledPetriNet,fname:str,
        include_prom_bits:bool=True
        ) -> None: 
    """
    Converts a given Petri net, to an XML structure conforming to the pnml 
    schema, then writes out the XML to a given file location (fname).

    No checking of file location is done for you.
    """
    xml =  convert_net_to_xml(net,include_prom_bits=include_prom_bits)  
    ET.indent( xml ) 
    ET.ElementTree(xml).write(fname,xml_declaration=True, encoding="utf-8")


"""
This module is a implementation of the alpha miner (original).

See the following article for more information about the 
alpha miner process discovery technique:\n
W.M.P. van der Aalst, A.J.M.M. Weijters, and L. Maruster. 
Workflow Mining: Discovering Process Models from Event Logs. 
IEEE Transactions on Knowledge and Data Engineering,
16(9):1128–1142, 2004.
"""
from typing import Set,Tuple,Dict,List,Union
from enum import Enum
from copy import deepcopy,copy

from pmkoalas.simple import EventLog
from pmkoalas._logging import info,debug
from pmkoalas.directly import DirectlyFollowPair
from pmkoalas.models.petrinets.pn import LabelledPetriNet, Arc
from pmkoalas.models.petrinets.pn import PetriNetMarking, AcceptingPetriNet
from pmkoalas.discovery.meta import DiscoveryTechnique

class AlphaRelationType(Enum):
    DF = ">"
    CD = "->"
    NF = "#"
    PD = "||"

class AlphaRelation():
    """
    A helper data class to work out what relation is suitable
    for a row, column in the footprint matrix for the alpha
    miner.

    Supported Relations:
    --------------------
    >  : directly follows\n
    -> : causal dependency\n
    \# : never follows\n
    || : parallel dependency\n
    """

    def __init__(self, src:str, target:str, follows:List[Tuple[str,str]]=None) -> None:
        self.target = target
        self.src = src
        self._follows = [] if follows == None else follows
        self.__members = [src, target]
        self.__hash = hash((target,src))

    def add_follows(self, src:str, follow:str) -> None:
        if (src in self.__members and follow in self.__members):
            self._follows.append((src,follow))

    def relation(self) -> AlphaRelationType:
        "returns the alpha relation between src and target"
        if ((self.src,self.target) in self._follows and 
            (self.target,self.src) in self._follows
        ):
            return AlphaRelationType.PD 
        elif ((self.src,self.target) in self._follows and 
              (self.target,self.src) not in self._follows
        ):
            return AlphaRelationType.CD 
        elif ( (self.src,self.target) in self._follows):
            return AlphaRelationType.DF
        return AlphaRelationType.NF

    # data model functions
    def __str__(self) -> str:
        formatter = "({src} {relation} {target})"
        return formatter.format(src=self.src, target=self.target,
            relation=self.relation().value
        )

    def __repr__(self) -> str:
        formatter = 'AlphaRelation(src="{src}",target="{target}",follows={follows})'
        return formatter.format(src=self.src, target=self.target,
            follows=repr(self._follows)
        )

    def __hash__(self) -> int:
        return self.__hash

    def __eq__(self, __o: object) -> bool:
        if (isinstance(__o, AlphaRelation)):
            return __o.__hash__() == self.__hash__()
        return False

class AlphaPair():
    """
    Data class helper for place generation.
    """

    def __init__(self, left:Set, right:Set) -> None:
        self.left = deepcopy(left) 
        self.right = deepcopy(right)

    def can_add_left(self, add:str, 
        matrix:Dict[str,Dict[str,AlphaRelation]]) -> bool:
        "Checks if we can add new input transition"
        # check that add is unseen
        if (add in self.left):
            return False
        # check for cd between new left and old rights
        for right in self.right:
            relation = matrix[add][right]
            if (relation.relation() != AlphaRelationType.CD):
                return False
        # check for nf between add and old lefts
        for left in self.left:
            relation = matrix[add][left]
            if (relation.relation() != AlphaRelationType.NF):
                return False
        return True

    def can_add_right(self, add:str, 
        matrix:Dict[str,Dict[str,AlphaRelation]]) -> bool:
        "Checks if we can add new output transition"
        # check that add is unseen
        if (add in self.right):
            return False
        # check for cd between old lefts to new right
        for left in self.left:
            relation = matrix[left][add]
            if (relation.relation() != AlphaRelationType.CD):
                return False
        # check for nf between add and old rights
        for right in self.right:
            relation = matrix[add][right]
            if (relation.relation() != AlphaRelationType.NF):
                return False
        return True

    def expand_left(self, add:str) -> 'AlphaPair':
        "Creates new pair with expanded left, using given add."
        new_left = deepcopy(self.left)
        new_left.add(add)
        new_right = deepcopy(self.right)
        return AlphaPair(new_left, new_right)

    def expand_right(self, add:str) -> 'AlphaPair':
        "Creates new pair with expanded right, using given add."
        new_left = deepcopy(self.left)
        new_right = deepcopy(self.right)
        new_right.add(add)
        return AlphaPair(new_left, new_right)

    def issubset(self, other:'AlphaPair') -> bool:
        "Checks if other implies this pair"
        if (not self.left.issubset(other.left)):
            return False 
        if (not self.right.issubset(other.right)):
            return False 
        return True

    def __hash__(self) -> int:
        ord_left = sorted(list([l for l in self.left]))
        ord_right = sorted(list([r for r in self.right]))
        return hash(
            tuple(["left:"]+[l for l in ord_left]+["right:"]+[r for r in ord_right])
        )

    def __repr__(self) -> str:
        ord_left = sorted([l for l in self.left])
        ord_right = sorted([r for r in self.right])
        left_str = "{" + str(ord_left)[1:-1] + "}"
        right_str = "{" + str(ord_right)[1:-1] + "}"
        return f"AlphaPair({left_str},{right_str})"
    
    def __str__(self) -> str:
        ord_left = sorted([l for l in self.left])
        ord_right = sorted([r for r in self.right])
        left_str = "{" + str(ord_left)[1:-1] + "}"
        right_str = "{" + str(ord_right)[1:-1] + "}"
        return f"({left_str},{right_str})"

    def __eq__(self, __o: object) -> bool:
        if (isinstance(__o, AlphaPair)):
            return __o.__hash__() == self.__hash__()
        return False

class AlphaPlace():
    """
    Data class helper for managing places generated from alpha miner.
    """

    def __init__(self, suffix:str, incoming:Set[str], 
        outgoing:Set[str]) -> None:
        self.name = "P_"+suffix
        self.incoming = deepcopy(incoming)
        self.outgoing = deepcopy(outgoing)
        self.__hash = hash(
            tuple(inc for inc in sorted(list(self.incoming)))
            + tuple(out for out in sorted(list(self.outgoing)))
        )
    
    def __hash__(self) -> int:
        return self.__hash

    def __eq__(self, __o: object) -> bool:
        if (isinstance(__o, AlphaPlace)):
            return __o.__hash__() == self.__hash__()
        return False 

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.__str__()

class AlphaStartPlace(AlphaPlace):
    """
    Special case for starting place in alpha miner.
    """

    ALPHA_START_HASH = "THISISTHEINITIALPLACE"

    def __init__(self) -> None:
        super().__init__("initial", set(), set())

    def __hash__(self) -> int:
        return hash(tuple(self.ALPHA_START_HASH))

class AlphaSinkPlace(AlphaPlace):
    """
    Special case for sink place in alpha miner.
    """

    ALPHA_END_HASH = "THISISTHESINKPLACE"

    def __init__(self) -> None:
        super().__init__("sink", set(), set())

    def __hash__(self) -> int:
        return hash(tuple(self.ALPHA_END_HASH))

class AlphaTransition():
    """
    Data class helper for managing transitions generated from
    alpha miner.
    """

    def __init__(self,name:str) -> None:
        self.name = name
        self.__hash = hash(tuple(name))

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return self.name

    def __hash__(self) -> int:
        return self.__hash

    def __eq__(self, __o: object) -> bool:
        if (isinstance(__o, AlphaTransition)):
            return self.__hash__() == __o.__hash__()
        return False
    
class AlphaFlowRelation():
    """
    Data class helper for managing flow relations generated
    from alpha miner.
    """

    def __init__(self,suffix:str, 
        src:Union[AlphaPlace,AlphaTransition],
        tar:Union[AlphaPlace,AlphaTransition]) -> None:
        self.name = "F_"+suffix
        self.src = src 
        self.tar = tar 
        self.__hash = hash(tuple([src,tar]))

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return f"{self.src} -> {self.tar}"

    def __hash__(self) -> int:
        return self.__hash

    def __eq__(self, __o: object) -> bool:
        if (isinstance(__o, AlphaFlowRelation)):
            return self.__hash__() == __o.__hash__()
        return False

class AlphaMinerInstance(DiscoveryTechnique):
    """
    This is a helper for the alpha miner using the same 
    parameters. The alpha miner consists of several steps,
    which each steps finds a finite set defined by an event 
    log. The alpha miner algorithm returns the components 
    needed to defined a Petri net, i.e. a set of places, a 
    set of transitions, and a set of flow relations.
    \n
    step one: find the set of transitions (process activities)\n
    step two: find the set of start activities\n
    step three: find the set of end activities\n
    step four: determine all possible places for the net\n
    step five: determine the maximal pairs for the net\n
    step six: \n
    step seven: \n
    \n
    The alpha miner has the following parameters:\n
    Parameters
    ----------    
    min_inst:`int=1`(default)\n
    The minimum number of times a relation must be seen to
    be counted as a alpha relation.
    """

    def __init__(self, min_inst:int=1, optimised:bool=False) -> None:
        self._min_inst = min_inst
        self._matrix = None
        self._opt = optimised

    def mine_footprint_matrix(self, log:EventLog) -> Dict[
        str,Dict[str,AlphaRelation]]:
        """
        Mines the footprint matrix of the alpha miner relations
        between process activities.
        """
        info("Mining footprint matrix...")
        self._matrix = dict()
        TL = self._step_one(log)
        # set up matrix 
        for src in TL:
            self._matrix[src] = dict()
            for target in TL:
                self._matrix[src][target] = AlphaRelation(
                    src=src,target=target
                )
        # find directly flows lang
        flang = log.directly_follow_relations()
        # update relations
        for col in self._matrix.values():
            for relation in col.values():
                src = relation.src
                target = relation.target
                # test flow relation from src to target
                pair = DirectlyFollowPair(src, target, 1)
                if (flang.contains(pair)):
                    fpair = flang.find(pair)
                    if fpair.frequency() >= self._min_inst:
                        # print(f"adding follows ({src},{target}) to {relation}")
                        relation.add_follows(src, target)
                # test flow relation from target to src
                pair = DirectlyFollowPair(target, src, 1)
                if (flang.contains(pair)):
                    fpair = flang.find(pair)
                    if fpair.frequency() >= self._min_inst:
                        # print(f"adding follows ({target},{src}) to {relation}")
                        relation.add_follows(target, src)
        # make new copy
        out = dict()
        for src in TL:
            out[src] = deepcopy(self._matrix[src])
        info("Footprint matrix mined.")
        return deepcopy(self._matrix)

    def discover(self, log:EventLog) -> AcceptingPetriNet:
        """
        Mines a petri net based on the given event log, using
        the alpha miner discovery technique.

        Returns a Petri net, i.e. a set of places, a set of 
        transitions and a set of flow relations.
        """
        self._matrix = None
        # find the essential elements 
        matrix = self.mine_footprint_matrix(log)
        ### the set of transitions
        info("Discovering Petri net...")
        info("Step one: finding transitions")
        TL = self._step_one(log)
        info("Step one: completed")
        # others 
        info("Step two: finding start activities")
        TI = self._step_two(log)
        info("Step two: completed")
        info("Step three: finding end activities")
        TO = self._step_three(log)
        info("Step three: completed")
        info("Step four: finding pairs")
        XL = self._step_four(log)
        info("Step four: completed")
        info("Step five: finding maximal pairs")
        YL = self._step_five(log, XL)
        info("Step five: completed")
        # find the petri net components 
        ### the set of places
        info("Step six: finding places")
        PL = self._step_six(log, YL) 
        info("Step six: completed")
        ### the set of flow relations
        info("Step seven: finding flow relations")
        FL = self._step_seven(log, PL, TI, TO) 
        info("Step seven: completed")
        info("Petri net discovered.")
        TL = set([ AlphaTransition(t) for t in TL])
        self._matrix = None
        info("Alpha Miner Returning Petri Net")
        from pmkoalas.models.petrinets.pn import Place, Transition, Arc
        places = dict()
        transitions = dict()
        for place in PL:
            places[place.name] = Place(place.name)
        for transition in TL:
            transitions[transition.name] = Transition(transition.name)
        net = LabelledPetriNet(
            set( p for p in places.values()), 
            set(t for t in transitions.values()),
            [Arc(
                places[f.src.name] if 
                isinstance(f.src, AlphaPlace) 
                else transitions[f.src.name],
                places[f.tar.name] if 
                isinstance(f.tar, AlphaPlace) 
                else transitions[f.tar.name],
            )
            for f in FL])
        imark = {}
        fmark = {}
        for place in net.places:
            inc_arcs = set(
                arc 
                for arc in net.arcs 
                if arc.to_node == place
            )
            out_arcs = set(
                arc 
                for arc in net.arcs 
                if arc.from_node == place
            )
            if len(inc_arcs) == 0:
                imark[place] = 1
            if len(out_arcs) == 0:
                fmark[place] = 1
        imark = PetriNetMarking(imark)
        fmark = PetriNetMarking(fmark)
        anet = AcceptingPetriNet(net, imark, [fmark])
        return anet

    def _step_one(self, log:EventLog) -> Set[str]:
        """
        Returns a set of all process activities seen in the 
        event log.
        """
        return log.seen_activities()

    def _step_two(self, log:EventLog) -> Set[str]:
        """
        Returns a set of all start process activities seen in 
        the event log.
        """
        return log.seen_start_activities()

    def _step_three(self, log:EventLog) -> Set[str]:
        """
        Returns a set of all end process activities seen in 
        the event log.
        """
        return log.seen_end_activities()

    def _step_four(self, log:EventLog) -> Set:
        """
        Returns a set of pairs, where each pair describes a 
        set of transitions that have a causal dependency (->) 
        with all the transitions in the second set. Futhermore, 
        each member of each set, is never seen follows (#) 
        another member from their set.
        """
        if (self._matrix == None):
            self._matrix = self.mine_footprint_matrix(log)
        pairs:Set[AlphaPair] = set()
        # build all pair iteratively, start within |A| = 1
        TL = self._step_one(log)
        for src in TL:
            col = self._matrix[src]
            # check that src has never follows with itself
            if (col[src].relation() != AlphaRelationType.NF):
                continue
            for tar in TL:
                relation = col[tar]
                # check for casual dependency
                if (relation.relation() == AlphaRelationType.CD):
                    # add to base
                    pairs.add( 
                        AlphaPair(set([src]),set([tar]))
                    )
        # keep expanding until pairs does not grow
        last_size = 0
        next_size = len(pairs)
        from joblib import Parallel, delayed
        pool = Parallel(n_jobs=-3,return_as='generator_unordered')
        while last_size != next_size:
            last_size = len(pairs)
            # to expand we do the following for each pair in pairs
            # we check for adding a new left for each pair
            # then check for adding a new right for each pair

            if (self._opt):
                ## optimised path
                # define work for workers
                def work(t:str, pairs:Set[AlphaPair], matrix) -> Set[AlphaPair]:
                    from pmkoalas.discovery.alpha_miner import AlphaPair
                    new_pairs = set()
                    pairs= eval(pairs)
                    for pair in pairs:
                        if (pair.can_add_left(t, matrix)):
                            gen = pair.expand_left(t)
                            if gen not in pairs:
                                new_pairs.add(gen)
                        if (pair.can_add_right(t, matrix)):
                            gen = pair.expand_right(t)
                            if gen not in pairs:
                                new_pairs.add(gen)
                    return repr(new_pairs)
                # expand pairs
                repred_pairs = repr(pairs)
                pool_of_pairs = pool(
                    delayed(work)(t, repred_pairs, self._matrix)
                    for t in TL
                )
                for sync_pairs in pool_of_pairs:
                    info("worker finished expanding...")
                    pairs.update(eval(sync_pairs))
            else:
                new_pairs = set()
                for t in TL:
                    for pair in pairs:
                        if (pair.can_add_left(t, self._matrix)):
                            new_pairs.add(pair.expand_left(t))
                        if (pair.can_add_right(t, self._matrix)):
                            new_pairs.add(pair.expand_right(t))
                pairs.update(new_pairs)
            # update end condition
            next_size = len(pairs)
            info(f"Pairs expanded this round: {next_size - last_size}...")
        return pairs

    def _step_five(self, log:EventLog, XL:Set) -> Set:
        """
        Returns all maximal pairs from XL, reducing the number
        of places in the Petri net.
        """
        maximal_pairs = set()
        for pair in XL:
            issubset = False
            for opair in XL:
                if pair == opair:
                    continue 
                issubset = issubset or pair.issubset(opair)
                if (issubset):
                    break
            if (not issubset):
                maximal_pairs.add(deepcopy(pair))
        return maximal_pairs 

    def _step_six(self, log:EventLog, YL:Set[AlphaPair]) -> Set:
        """
        Returns a set of alpha places for the Petri net, including 
        an initial place and a sink place.
        """
        places = set()
        for y in YL:
            places.add(AlphaPlace(str(y), y.left, y.right))
        places.add(AlphaStartPlace())
        places.add(AlphaSinkPlace())
        return places

    def _step_seven(self, log:EventLog, PL:Set[AlphaPlace], 
        TI:Set[str], TO:Set[str]) -> Set:
        """
        Returns a set of alpha flow relations, which are directed
        arcs from source to some target.
        """
        flows = set()
        for place in PL:
            if isinstance(place, AlphaSinkPlace):
                for out in TO:
                    flows.add(AlphaFlowRelation(str(len(flows)+1), 
                        AlphaTransition(out),
                        place
                    )) 
            elif isinstance(place, AlphaStartPlace):
                for inc in TI:
                    flows.add(AlphaFlowRelation(str(len(flows)+1), 
                        place,
                        AlphaTransition(inc)
                    )) 
            else:
                for inc in place.incoming:
                    flows.add(AlphaFlowRelation(str(len(flows)+1),
                        AlphaTransition(inc),
                        place 
                    ))
                for out in place.outgoing:
                    flows.add(AlphaFlowRelation(str(len(flows)+1),
                        place, 
                        AlphaTransition(out)
                    ))
        return flows

### ------------------- Alpha Miner Plus ------------------- ###

class AlphaPlusRelation(AlphaRelation):
    """
    A helper class to determine the ordering relations that captures
    length-two loops.

    Follows the work of the following article: Alves De Medeiros, A. K., 
    Dongen, van, B. F., Aalst, van der, W. M. P., & Weijters, A. J. M. M. 
    (2004). Process mining : extending the alpha-algorithm to mine short 
    loops.

    The ordering relations are described in Defintion 3.3. of the article.
    """

    def __init__(self, src: str, target: str, follows: List[Tuple[str]] = None) -> None:
        super().__init__(src, target, follows)
        self._two_step_follows = set()
    
    def add_two_step_follows(self, src: str, follow: str) -> None:
        if (src in self._AlphaRelation__members):
            self._two_step_follows.add((src, follow))

    def relation(self) -> AlphaRelationType:
        # check for length two loop relations e.g. aba or bab 
        # check for ab-? then check for aba 
        loop_delta_a = False
        if ((self.src,self.target) in self._follows):
            if (self.target,self.src) in self._two_step_follows:
                loop_delta_a = True
        # check for ba-? then check for bab 
        loop_delta_b = False
        if ((self.target,self.src) in self._follows):
            if (self.src,self.target) in self._two_step_follows:
                loop_delta_b = True
        
        looping = loop_delta_b and loop_delta_a
        # return one of the original alpha relations
        if ((self.src,self.target) in self._follows and 
            (self.target,self.src) in self._follows and
            not looping
        ):
            return AlphaRelationType.PD 
        elif ((self.src,self.target) in self._follows and 
              (
                    (self.target,self.src) not in self._follows
                    or 
                    looping
             )
        ):
            return AlphaRelationType.CD 
        elif ( (self.src,self.target) in self._follows):
            return AlphaRelationType.DF
        return AlphaRelationType.NF

class AlphaMinerPlusInstance(DiscoveryTechnique):
    """
    Mines a Petri net using the alpha miner plus algorithm as presented in
    Alves De Medeiros, A. K., 
    Dongen, van, B. F., Aalst, van der, W. M. P., & Weijters, A. J. M. M. 
    (2004). Process mining : extending the alpha-algorithm to mine short 
    loops.

    The alpha miner plus algorithm is an extension of the alpha miner,
    whereby preprocessing the given log and adjusting the ordering relations
    the alpha-plus miner can mine petri nets with short loops.
    """

    def __init__(self, min_inst:int=1 ,optimised:bool=False) -> None:
        super().__init__()
        self._opt = optimised
        self._min_inst = min_inst

    def mine_footprint_matrix(self, log:EventLog) -> Dict[
        str,Dict[str,AlphaPlusRelation]]:
        """
        Mines the footprint matrix of the alpha miner relations
        between process activities.
        """
        info("Mining footprint matrix...")
        self._matrix = dict()
        TL = self._step_one(log)
        # set up matrix 
        for src in TL:
            self._matrix[src] = dict()
            for target in TL:
                self._matrix[src][target] = AlphaPlusRelation(
                    src=src,target=target
                )
        # find directly flows lang
        flang = log.directly_follow_relations()
        # update relations
        for col in self._matrix.values():
            for relation in col.values():
                src = relation.src
                target = relation.target
                # test flow relation from src to target
                pair = DirectlyFollowPair(src, target, 1)
                if (flang.contains(pair)):
                    fpair = flang.find(pair)
                    if fpair.frequency() >= self._min_inst:
                        # print(f"adding follows ({src},{target}) to {relation}")
                        relation.add_follows(src, target)
                        # check for aba in proceeding
                        if src in fpair.proceeding():
                            relation.add_two_step_follows(src, target)
                # test flow relation from target to src
                pair = DirectlyFollowPair(target, src, 1)
                if (flang.contains(pair)):
                    fpair = flang.find(pair)
                    if fpair.frequency() >= self._min_inst:
                        # print(f"adding follows ({target},{src}) to {relation}")
                        relation.add_follows(target, src)
                        # check for bab in proceeding
                        if target in fpair.proceeding():
                            relation.add_two_step_follows(target, src)
        # make new copy
        out = dict()
        for src in TL:
            out[src] = deepcopy(self._matrix[src])
        info("Footprint matrix mined.")
        return deepcopy(self._matrix)
    
    def discover(self, log:EventLog) -> LabelledPetriNet:
        """
        Computes a Petri net to represent the given event log, assuming that
        the event log is a loop-complete workflow log (def 3.1.).

        The algorithm consists of nine steps where the alpha miner alogrithm
        is called but mines ordering relations that capture length-two loops,
        as described in Def. 4.4 of the article.
        """
        from pmkoalas.models.petrinets.pn import Transition,Arc, Place
        self._matrix = None
        _ = self.mine_footprint_matrix(log)
        #step one 
        acts = self._step_one(log)
        #step two 
        doubles = self._step_two(log)
        #step three 
        nots = self._step_three(acts, doubles)
        #step four 
        fdoubles = self._step_four(doubles, nots)
        #step five
        wlog = self._step_five(log, doubles)
        #step six
        net = self._step_six(wlog)
        # return net 
        places = net.places 
        transitions = net.transitions
        ntransitions = set()
        # the ids of the places need to match with the transitions
        # for fun downstream stuff...
        swaps = dict()
        missing = set()
        for f in fdoubles:
            if isinstance(f.from_node,Place) and f.from_node.name not in swaps:
                options = set( 
                    p 
                    for p in places 
                    if p.name == f.from_node.name
                )
                if (len(options) > 0):
                    swaps[f.from_node.name] = options.pop()
                else:
                    missing.add(f.from_node)
            elif isinstance(f.from_node, Transition):
                ntransitions.add(
                    f.from_node
                )
            if isinstance(f.to_node,Place) and f.to_node not in swaps:
                options = set( 
                    p 
                    for p in places 
                    if p.name == f.to_node.name
                )
                if (len(options) > 0):
                    swaps[f.to_node.name] = options.pop()
                else:
                    missing.add(f.to_node)
            elif isinstance(f.to_node, Transition):
                ntransitions.add(
                    f.to_node
                )
        nfdoubles = set()
        for f in fdoubles:
            nfdoubles.add(
                Arc(
                    swaps[f.from_node.name] if f.from_node.name in swaps.keys() 
                    else f.from_node,
                    swaps[f.to_node.name] if f.to_node.name in swaps.keys()
                    else f.to_node
                )
            )
        # now we union the flows
        transitions = transitions.union(ntransitions)
        flows = net.arcs.union(nfdoubles)
        places = places.union(missing)
        net = LabelledPetriNet(
            places, transitions, flows
        )
        imark = {}
        fmark = {}
        for place in net.places:
            inc_arcs = set(
                arc 
                for arc in net.arcs 
                if arc.to_node == place
            )
            out_arcs = set(
                arc 
                for arc in net.arcs 
                if arc.from_node == place
            )
            if len(inc_arcs) == 0:
                imark[place] = 1
            if len(out_arcs) == 0:
                fmark[place] = 1
        net = AcceptingPetriNet(
            net,
            PetriNetMarking(imark),
            [PetriNetMarking(fmark)]
        )
        return net 
    
    def _step_one(self, log:EventLog) -> Set[str]:
        """
        Returns a set of all process activities seen in the 
        event log.
        """
        return log.seen_activities()
    
    def _step_two(self, log:EventLog) -> Set[str]:
        """
        Returns a set of process activities seen in the event, such that there
        exists a trace in the log where ...aa... is seen.
        """
        doubles = set()
        for trace in log.language():
            for i in range(1,len(trace)):
                if (trace[i-1] == trace[i]):
                    doubles.add(trace[i])
        return doubles
    
    def _step_three(self, acts: Set[str], doubles: Set[str]) -> Set[str]:
        """
        Simply returns the set of activities that are not in one-length loops.
        """
        return acts.difference(doubles)
    
    def _step_four(self, doubles:Set[str], nots: set[str]) -> Set['Arc']:
        """
        Returns a set of flow relations that capture the ordering relations
        that capture length-one loops.
        """
        from pmkoalas.models.petrinets.pn import Place,Transition,Arc
        flows = set()
        for t in doubles:
            places_A = set(
                a 
                for a in nots 
                if self._matrix[a][t].relation() == AlphaRelationType.CD
            )
            places_B = set(
                b 
                for b in nots 
                if self._matrix[t][b].relation() == AlphaRelationType.CD
            )
            trans = Transition(t)
            if len(places_A) == 0 and len(places_B) == 0:
                place = AlphaStartPlace()
            else:
                pair = AlphaPair(places_A.difference(places_B), 
                                places_B.difference(places_A)
                )
                place = AlphaPlace(str(pair), pair.left, pair.right)
            place = Place(place.name)
            flows.add(Arc(trans, place))
            flows.add(Arc(place, trans))
        return flows
    
    def _step_five(self, log:EventLog, doubles: set[str]) -> EventLog:
        """
        Returns a new event log where all traces have been processed to remove
        activities that were identified in one-length loops.
        """
        from pmkoalas.simple import Trace
        new_traces = []
        for trace, freq in log.__iter__():
            new_trace = []
            for act in trace:
                if act not in doubles:
                    new_trace.append(act)
            for i in range(freq):
                new_traces.append(Trace(new_trace))
        return EventLog(new_traces)
    
    def _step_six(self, log:EventLog) -> LabelledPetriNet:
        """
        Returns the Petri net discovered by the alpha miner using the alpha
        plus relations.
        """
        miner = AlphaMinerInstance(min_inst=self._min_inst,optimised=self._opt)
        miner.mine_footprint_matrix = self.mine_footprint_matrix
        return miner.discover(log).net
        
    

